from typing import Optional, Union

from ufl import exp

from festim import k_B as _k_B
from festim.species import ImplicitSpecies as _ImplicitSpecies
from festim.species import Species as _Species
from festim.subdomain.volume_subdomain_1d import VolumeSubdomain1D as VS1D


class Reaction:
    """A reaction between two species, with a forward and backward rate.

    Arguments:
        reactant (Union[F.Species, F.ImplicitSpecies], List[Union[F.Species, F.ImplicitSpecies]]): The reactant.
        product (Optional[Union[F.Species, List[F.Species]]]): The product.
        k_0 (float): The forward rate constant pre-exponential factor.
        E_k (float): The forward rate constant activation energy.
        p_0 (float): The backward rate constant pre-exponential factor.
        E_p (float): The backward rate constant activation energy.
        volume (F.VolumeSubdomain1D): The volume subdomain where the reaction takes place.

    Attributes:
        reactant (Union[F.Species, F.ImplicitSpecies], List[Union[F.Species, F.ImplicitSpecies]]): The reactant.
        product (Optional[Union[F.Species, List[F.Species]]]): The product.
        k_0 (float): The forward rate constant pre-exponential factor.
        E_k (float): The forward rate constant activation energy.
        p_0 (float): The backward rate constant pre-exponential factor.
        E_p (float): The backward rate constant activation energy.
        volume (F.VolumeSubdomain1D): The volume subdomain where the reaction takes place.

    Usage:
        >>> # create two species
        >>> reactant = [F.Species("A"), F.Species("B")]

        >>> # create a product species
        >>> product = F.Species("C")

        >>> # create a reaction between the two species
        >>> reaction = Reaction(reactant, product, k_0=1.0, E_k=0.2, p_0=0.1, E_p=0.3)
        >>> print(reaction)
        A + B <--> C

        >>> # compute the reaction term at a given temperature
        >>> temperature = 300.0
        >>> reaction_term = reaction.reaction_term(temperature)

    """

    def __init__(
        self,
        reactant: _Species | _ImplicitSpecies | list[_Species | _ImplicitSpecies],
        k_0: float,
        E_k: float,
        volume: VS1D,
        product: Optional[Union[_Species, list[_Species]]] = [],
        p_0: float = None,
        E_p: float = None,
    ) -> None:
        self.k_0 = k_0
        self.E_k = E_k
        self.p_0 = p_0
        self.E_p = E_p
        self.reactant = reactant
        self.product = product
        self.volume = volume

    @property
    def reactant(self):
        return self._reactant

    @reactant.setter
    def reactant(self, value):
        if not isinstance(value, list):
            value = [value]
        if len(value) == 0:
            raise ValueError(
                "reactant must be an entry of one or more species objects, not an empty list."
            )
        for i in value:
            if not isinstance(i, (_Species, _ImplicitSpecies)):
                raise TypeError(
                    "reactant must be an F.Species or F.ImplicitSpecies, not "
                    + f"{type(i)}"
                )
        self._reactant = value

    def __repr__(self) -> str:
        reactants = " + ".join([str(reactant) for reactant in self.reactant])

        if isinstance(self.product, list):
            products = " + ".join([str(product) for product in self.product])
        else:
            products = self.product
        return f"Reaction({reactants} <--> {products}, {self.k_0}, {self.E_k}, {self.p_0}, {self.E_p})"

    def __str__(self) -> str:
        reactants = " + ".join([str(reactant) for reactant in self.reactant])
        if isinstance(self.product, list):
            products = " + ".join([str(product) for product in self.product])
        else:
            products = self.product
        return f"{reactants} <--> {products}"

    def reaction_term(self, temperature):
        """Compute the reaction term at a given temperature.

        Arguments:
            temperature (): The temperature at which the reaction term is computed.
        """

        if self.product == []:
            if self.p_0 is not None:
                raise ValueError(
                    f"p_0 must be None, not {self.p_0}"
                    + " when no products are present."
                )
            if self.E_p is not None:
                raise ValueError(
                    f"E_p must be None, not {self.E_p}"
                    + " when no products are present."
                )
        else:
            if self.p_0 == None:
                raise ValueError(
                    "p_0 cannot be None when reaction products are present."
                )
            elif self.E_p == None:
                raise ValueError(
                    "E_p cannot be None when reaction products are present."
                )

        k = self.k_0 * exp(-self.E_k / (_k_B * temperature))

        if self.p_0 and self.E_p:
            p = self.p_0 * exp(-self.E_p / (_k_B * temperature))
        elif self.p_0:
            p = self.p_0
        else:
            p = 0

        reactants = self.reactant
        product_of_reactants = reactants[0].concentration
        for reactant in reactants[1:]:
            product_of_reactants *= reactant.concentration

        if isinstance(self.product, list):
            products = self.product
        else:
            products = [self.product]

        if len(products) > 0:
            product_of_products = products[0].solution
            for product in products[1:]:
                product_of_products *= product.solution
        else:
            product_of_products = 0
        return k * product_of_reactants - (p * product_of_products)



from mpi4py import MPI

from dolfinx.io import XDMFFile

import festim as F


class MeshFromXDMF(F.Mesh):
    """
    Mesh read from the XDMF files

    Args:
        volume_file (str): path to the volume file
        facet_file (str): path to the facet file
        mesh_name (str, optional): name of the mesh in the volume XDMF file.
            Defaults to "Grid".
        surface_meshtags_name (str, optional): name of the surface meshtags in the
            facet XDMF file. Defaults to "Grid".
        volume_meshtags_name (str, optional): name of the volume meshtags in the
            volume XDMF file. Defaults to "Grid".

    Attributes:
        volume_file (str): path to the volume file
        facet_file (str): path to the facet file
        mesh_name (str): name of the mesh in the volume XDMF file.
        surface_meshtags_name (str): name of the surface meshtags in the facet XDMF file.
        volume_meshtags_name (str): name of the volume meshtags in the volume XDMF file
        mesh (fenics.mesh.Mesh): the fenics mesh
    """

    def __init__(
        self,
        volume_file,
        facet_file,
        mesh_name="Grid",
        surface_meshtags_name="Grid",
        volume_meshtags_name="Grid",
    ) -> None:
        self.volume_file = volume_file
        self.facet_file = facet_file
        self.mesh_name = mesh_name
        self.surface_meshtags_name = surface_meshtags_name
        self.volume_meshtags_name = volume_meshtags_name

        volumes_file = XDMFFile(MPI.COMM_WORLD, self.volume_file, "r")
        mesh = volumes_file.read_mesh(name=f"{self.mesh_name}")

        super().__init__(mesh=mesh)

    def define_surface_meshtags(self):
        """Creates the facet meshtags

        Returns:
            dolfinx.MeshTags: the facet meshtags
        """
        facets_file = XDMFFile(MPI.COMM_WORLD, self.facet_file, "r")
        facet_meshtags = facets_file.read_meshtags(
            self.mesh, name=f"{self.surface_meshtags_name}"
        )

        return facet_meshtags

    def define_volume_meshtags(self):
        """Creates the volume meshtags

        Returns:
            dolfinx.MeshTags: the volume meshtags
        """
        volume_file = XDMFFile(MPI.COMM_WORLD, self.volume_file, "r")

        volume_meshtags = volume_file.read_meshtags(
            self.mesh, name=f"{self.volume_meshtags_name}"
        )

        return volume_meshtags



import numpy as np

from festim.exports import VolumeQuantity


class MaximumVolume(VolumeQuantity):
    """Computes the maximum value of a field in a given volume

    Args:
        field (festim.Species): species for which the maximum volume is computed
        volume (festim.VolumeSubdomain): volume subdomain
        filename (str, optional): name of the file to which the maximum volume is exported

    Attributes:
        see `festim.VolumeQuantity`

    """

    @property
    def title(self):
        return f"Maximum {self.field.name} volume {self.volume.id}"

    def compute(self):
        """
        Computes the maximum value of solution function within the defined volume
        subdomain, and appends it to the data list
        """
        solution = self.field.solution
        indices = self.volume.locate_subdomain_entities(solution.function_space.mesh)
        self.value = np.max(self.field.solution.x.array[indices])
        self.data.append(self.value)



import mpi4py.MPI as MPI

import numpy as np
import ufl
from dolfinx import fem


def error_L2(u_computed, u_exact, degree_raise=3):
    # Create higher order function space
    degree = u_computed.function_space.ufl_element().degree
    family = u_computed.function_space.ufl_element().family_name
    mesh = u_computed.function_space.mesh
    W = fem.functionspace(mesh, (family, degree + degree_raise))
    # Interpolate approximate solution
    u_W = fem.Function(W)
    u_W.interpolate(u_computed)

    # Interpolate exact solution, special handling if exact solution
    # is a ufl expression or a python lambda function
    u_ex_W = fem.Function(W)
    if isinstance(u_exact, ufl.core.expr.Expr):
        u_expr = fem.Expression(u_exact, W.element.interpolation_points)
        u_ex_W.interpolate(u_expr)
    else:
        u_ex_W.interpolate(u_exact)

    # Compute the error in the higher order function space
    e_W = fem.Function(W)
    e_W.x.array[:] = u_W.x.array - u_ex_W.x.array

    # Integrate the error
    error = fem.form(ufl.inner(e_W, e_W) * ufl.dx)
    error_local = fem.assemble_scalar(error)
    error_global = mesh.comm.allreduce(error_local, op=MPI.SUM)
    return np.sqrt(error_global)



import dolfinx
import numpy as np


def transfer_meshtags_to_submesh(
    mesh, entity_tag, submesh, sub_vertex_to_parent, sub_cell_to_parent
):
    """
    Transfer a meshtag from a parent mesh to a sub-mesh.
    """

    # Invert map from sub-mesh to parent-mesh to parent-mesh to sub-mesh
    tdim = mesh.topology.dim
    cell_imap = mesh.topology.index_map(tdim)
    num_cells = cell_imap.size_local + cell_imap.num_ghosts
    mesh_to_submesh = np.full(num_cells, -1)
    mesh_to_submesh[sub_cell_to_parent] = np.arange(
        len(sub_cell_to_parent), dtype=np.int32
    )

    # Compute and access various entity to vertex and vertex to
    # entity connectivity for parent and sub-mesh.
    submesh.topology.create_connectivity(entity_tag.dim, 0)

    num_child_entities = (
        submesh.topology.index_map(entity_tag.dim).size_local
        + submesh.topology.index_map(entity_tag.dim).num_ghosts
    )
    submesh.topology.create_connectivity(submesh.topology.dim, entity_tag.dim)

    c_c_to_e = submesh.topology.connectivity(submesh.topology.dim, entity_tag.dim)
    c_e_to_v = submesh.topology.connectivity(entity_tag.dim, 0)

    child_markers = np.full(num_child_entities, 0, dtype=np.int32)

    mesh.topology.create_connectivity(entity_tag.dim, 0)
    mesh.topology.create_connectivity(entity_tag.dim, mesh.topology.dim)
    p_f_to_v = mesh.topology.connectivity(entity_tag.dim, 0)
    p_f_to_c = mesh.topology.connectivity(entity_tag.dim, mesh.topology.dim)
    sub_to_parent_entity_map = np.full(num_child_entities, -1, dtype=np.int32)
    for facet, value in zip(entity_tag.indices, entity_tag.values):
        facet_found = False
        for cell in p_f_to_c.links(facet):
            if facet_found:
                break
            if (child_cell := mesh_to_submesh[cell]) != -1:
                for child_facet in c_c_to_e.links(child_cell):
                    child_vertices = c_e_to_v.links(child_facet)
                    child_vertices_as_parent = sub_vertex_to_parent[child_vertices]
                    is_facet = np.isin(
                        child_vertices_as_parent, p_f_to_v.links(facet)
                    ).all()
                    if is_facet:
                        child_markers[child_facet] = value
                        facet_found = True
                        sub_to_parent_entity_map[child_facet] = facet
    tags = dolfinx.mesh.meshtags(
        submesh,
        entity_tag.dim,
        np.arange(num_child_entities, dtype=np.int32),
        child_markers,
    )
    tags.name = entity_tag.name
    return tags, sub_to_parent_entity_map



import dolfinx
import numpy as np
from dolfinx.cpp.fem import compute_integration_domains

from .volume_subdomain import VolumeSubdomain


class Interface:
    id: int
    subdomains: tuple[VolumeSubdomain, VolumeSubdomain]
    parent_mesh: dolfinx.mesh.Mesh
    mt: dolfinx.mesh.MeshTags
    restriction: list[str, str] = ("+", "-")
    padded: bool

    def __init__(
        self,
        id: int,
        subdomains: list[VolumeSubdomain],
        penalty_term: float = 10.0,
    ):
        """Class representing an interface between two subdomains.

        Args:
            id (int): the tag of the interface subdomain in the parent meshtags
            subdomains (list[F.VolumeSubdomain]): the subdomains sharing this interface
            penalty_term (float, optional): Penalty term in the Nietsche DG formulation.
                Needs to be "sufficiently large". Defaults to 10.0.
        """
        self.id = id
        self.subdomains = tuple(subdomains)
        self.penalty_term = penalty_term

    def pad_parent_maps(self):
        """Workaround to make sparsity-pattern work without skips"""

        integration_data = compute_integration_domains(
            dolfinx.fem.IntegralType.interior_facet,
            self.parent_mesh.topology._cpp_object,
            self.mt.find(self.id),
            self.mt.dim,
        ).reshape(-1, 4)
        for i in range(2):
            # We pad the parent to submesh map to make sure that sparsity pattern is correct
            mapped_cell_0 = self.subdomains[i].parent_to_submesh[integration_data[:, 0]]
            mapped_cell_1 = self.subdomains[i].parent_to_submesh[integration_data[:, 2]]
            max_cells = np.maximum(mapped_cell_0, mapped_cell_1)
            self.subdomains[i].parent_to_submesh[integration_data[:, 0]] = max_cells
            self.subdomains[i].parent_to_submesh[integration_data[:, 2]] = max_cells
            self.subdomains[i].padded = True

    def compute_mapped_interior_facet_data(self, mesh):
        """
        Compute integration data for interface integrals.
        We define the first domain on an interface as the "+" restriction,
        meaning that we must sort all integration entities in this order

        Returns
            integration_data: Integration data for interior facets
        """
        assert (not self.subdomains[0].padded) and (not self.subdomains[1].padded)
        mesh.topology.create_connectivity(mesh.topology.dim - 1, mesh.topology.dim)
        integration_data = compute_integration_domains(
            dolfinx.fem.IntegralType.interior_facet,
            mesh.topology._cpp_object,
            self.mt.find(self.id),
            self.mt.dim,
        )

        ordered_integration_data = integration_data.reshape(-1, 4).copy()

        mapped_cell_0 = self.subdomains[0].parent_to_submesh[integration_data[0::4]]
        mapped_cell_1 = self.subdomains[0].parent_to_submesh[integration_data[2::4]]

        switch = mapped_cell_1 > mapped_cell_0
        # Order restriction on one side
        if True in switch:
            ordered_integration_data[switch, :] = ordered_integration_data[switch][
                :, [2, 3, 0, 1]
            ]

        # Check that other restriction lies in other interface
        domain1_cell = self.subdomains[1].parent_to_submesh[
            ordered_integration_data[:, 2]
        ]
        assert (domain1_cell >= 0).all()

        return (self.id, ordered_integration_data.reshape(-1))



import numpy as np
from dolfinx import fem

import festim as F


def test_minimum_surface_export_compute_1D():
    """Test that the minimum surface export computes the correct value"""

    # BUILD
    L = 4.0
    D = 1.5
    my_mesh = F.Mesh1D(np.linspace(0, L, 10000))
    dummy_surface = F.SurfaceSubdomain1D(id=1, x=4)
    dummy_surface.locate_boundary_facet_indices(mesh=my_mesh.mesh)

    # give function to species
    V = fem.functionspace(my_mesh.mesh, ("Lagrange", 1))
    c = fem.Function(V)
    c.interpolate(lambda x: (x[0] - 2) ** 2)

    my_species = F.Species("H")
    my_species.solution = c

    my_export = F.MinimumSurface(field=my_species, surface=dummy_surface)
    my_export.D = D

    # RUN
    my_export.compute()

    # TEST
    expected_value = 4.0
    computed_value = my_export.value

    assert np.isclose(computed_value, expected_value, rtol=1e-2)



from mpi4py import MPI

import dolfinx
import numpy as np
import pytest
import ufl
from dolfinx.fem import Function, functionspace
from dolfinx.mesh import create_unit_cube

import festim as F


def test_assign_functions_to_species():
    """Test that checks if the function assign_functions_to_species
    creates the correct attributes for each species
    """

    mesh = F.Mesh1D(
        vertices=np.array([0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0])
    )
    model = F.HydrogenTransportProblem(
        mesh=mesh,
        species=[
            F.Species(name="H"),
            # F.Species(name="Trap"),
        ],
    )
    model.define_function_spaces()
    model.assign_functions_to_species()

    for spe in model.species:
        assert spe.solution is not None
        assert spe.prev_solution is not None
        assert spe.test_function is not None
        assert isinstance(spe.solution, dolfinx.fem.Function)
        assert isinstance(spe.prev_solution, dolfinx.fem.Function)
        assert isinstance(spe.test_function, ufl.Argument)


def test_species_repr_and_str():
    """Test that the __repr__ and __str__ methods of the Species class returns the
    expected string.
    """
    # create a species
    species = F.Species("A")

    # check that the __repr__ method returns the expected string
    expected_repr = "Species(A)"
    assert repr(species) == expected_repr

    # check that the __str__ method returns the expected string
    expected_str = "A"
    assert str(species) == expected_str


def test_implicit_species_repr_and_str():
    """Test that the __repr__ and __str__ methods of the ImplicitSpecies class
    returns the expected string.
    """
    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")

    # create an implicit species that depends on the two species
    implicit_species = F.ImplicitSpecies(3.0, [species1, species2], name="C")

    # check that the __repr__ method returns the expected string
    expected_repr = f"ImplicitSpecies(C, 3.0, {[species1, species2]})"
    assert repr(implicit_species) == expected_repr

    # check that the __str__ method returns the expected string
    expected_str = "C"
    assert str(implicit_species) == expected_str


def test_implicit_species_concentration():
    """Test that the concentration of an implicit species is computed
    correctly.
    """
    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")

    # create an implicit species that depends on the two species
    implicit_species = F.ImplicitSpecies(3.0, [species1, species2], name="C")

    # set the solutions of the two species
    mesh = create_unit_cube(MPI.COMM_WORLD, 10, 10, 10)
    V = functionspace(mesh, ("Lagrange", 1))
    species1.solution = Function(V)
    species2.solution = Function(V)

    implicit_species.create_value_fenics(mesh, t=0.0)

    # test the concentration of the implicit species
    expected_concentration = implicit_species.value_fenics - (
        species1.solution + species2.solution
    )
    assert implicit_species.concentration == expected_concentration


def test_implicit_species_concentration_with_no_solution():
    """Test that a ValueError is raised when on of the 'others' species
    has no solution and the concentration of the implicit species is
    requested.
    """
    # create two species
    species1 = F.Species("A")
    species2 = F.Species("B")

    # create an implicit species that depends on the two species
    implicit_species = F.ImplicitSpecies(3.0, [species1, species2], name="C")

    # set the solution of the first species
    mesh = create_unit_cube(MPI.COMM_WORLD, 10, 10, 10)
    V = functionspace(mesh, ("Lagrange", 1))
    species1.solution = Function(V)

    # test that a ValueError is raised when the second species has no solution
    with pytest.raises(ValueError):
        implicit_species.concentration


def test_create_species_and_reaction():
    """test that the trapped_concentration and trap_reaction attributes
    are correctly set"""

    # BUILD
    my_mobile_species = F.Species("test_mobile")
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=None)
    my_trap = F.Trap(
        name="test",
        mobile_species=my_mobile_species,
        k_0=1,
        E_k=1,
        p_0=1,
        E_p=1,
        n=1,
        volume=my_vol,
    )

    # RUN
    my_trap.create_species_and_reaction()

    # TEST
    assert isinstance(my_trap.trapped_concentration, F.Species)
    assert isinstance(my_trap.reaction, F.Reaction)


def test_implicit_species_wrong_type():
    """Test that a ValueError is raised when the value of n in an ImplicitSpecies is
    a function of time only but doesn't return a float or an int.
    """
    mesh = create_unit_cube(MPI.COMM_WORLD, 10, 10, 10)

    density = lambda t: "coucou"
    species = F.ImplicitSpecies(n=density)

    with pytest.raises(
        ValueError, match="self.value should return a float or an int, not "
    ):
        species.create_value_fenics(mesh=mesh, t=0.0)



import numpy as np
import pytest

import festim as F

test_mesh = F.Mesh1D(vertices=np.array([0.0, 1.0, 2.0, 3.0, 4.0]))


def test_define_diffusion_coefficient():
    """Test that the diffusion coefficient is correctly defined"""
    T, D_0, E_D = 10, 1.2, 0.5
    dum_spe = F.Species("dummy")
    my_mat = F.Material(D_0=D_0, E_D=E_D)
    D = my_mat.get_diffusion_coefficient(test_mesh.mesh, T, species=dum_spe)

    D_analytical = D_0 * np.exp(-E_D / F.k_B / T)

    assert np.isclose(float(D), D_analytical)


def test_multispecies_dict_strings():
    """Test that the diffusion coefficient is correctly defined when keys are
    strings"""
    T = 500
    D_0_A, D_0_B = 1, 2
    E_D_A, E_D_B = 0.1, 0.2
    A, B = F.Species("A"), F.Species("B")

    my_mat = F.Material(D_0={"A": D_0_A, "B": D_0_B}, E_D={"A": E_D_A, "B": E_D_B})
    D_A = my_mat.get_diffusion_coefficient(test_mesh.mesh, T, species=A)
    D_B = my_mat.get_diffusion_coefficient(test_mesh.mesh, T, species=B)

    D = [float(D_A), float(D_B)]

    D_A_analytical = D_0_A * np.exp(-E_D_A / F.k_B / T)
    D_B_analytical = D_0_B * np.exp(-E_D_B / F.k_B / T)

    D_analytical = [D_A_analytical, D_B_analytical]

    assert np.isclose(D, D_analytical).all()


def test_multispecies_dict_objects():
    """Test that the diffusion coefficient is correctly defined when keys are
    festim.Species objects"""
    T = 500
    D_0_A, D_0_B = 1, 2
    E_D_A, E_D_B = 0.1, 0.2

    A = F.Species("A")
    B = F.Species("B")

    my_mat = F.Material(D_0={A: D_0_A, B: D_0_B}, E_D={A: E_D_A, B: E_D_B})
    D_A = my_mat.get_diffusion_coefficient(test_mesh.mesh, T, species=A)
    D_B = my_mat.get_diffusion_coefficient(test_mesh.mesh, T, species=B)
    D = [float(D_A), float(D_B)]

    D_A_analytical = D_0_A * np.exp(-E_D_A / F.k_B / T)
    D_B_analytical = D_0_B * np.exp(-E_D_B / F.k_B / T)

    D_analytical = [D_A_analytical, D_B_analytical]

    assert np.isclose(D, D_analytical).all()


def test_multispecies_dict_objects_and_strings():
    """Test that the diffusion coefficient is correctly defined when keys
    are a mix of festim.Species objects and strings"""
    T = 500
    D_0_A, D_0_B = 1, 2
    E_D_A, E_D_B = 0.1, 0.2

    A = F.Species("A")
    B = F.Species("B")

    my_mat = F.Material(D_0={A: D_0_A, "B": D_0_B}, E_D={A: E_D_A, "B": E_D_B})
    D_A = my_mat.get_diffusion_coefficient(test_mesh.mesh, T, species=A)
    D_B = my_mat.get_diffusion_coefficient(test_mesh.mesh, T, species=B)
    D = [float(D_A), float(D_B)]

    D_A_analytical = D_0_A * np.exp(-E_D_A / F.k_B / T)
    D_B_analytical = D_0_B * np.exp(-E_D_B / F.k_B / T)

    D_analytical = [D_A_analytical, D_B_analytical]

    assert np.isclose(D, D_analytical).all()


def test_multispecies_dict_different_keys():
    """Test that a value error is raised when the keys of the D_0 and E_D
    are not the same"""
    A = F.Species("A")
    my_mat = F.Material(D_0={"A": 1, "B": 2}, E_D={"A": 0.1, "B": 0.2, "C": 0.3})

    with pytest.raises(ValueError, match="D_0 and E_D have different keys"):
        my_mat.get_diffusion_coefficient(test_mesh.mesh, 500, species=A)


def test_D_0_type_raises_error():
    """Test that a value error is raised in the get_diffusion_coefficient
    function"""
    # TODO remove this when material class is updated
    A = F.Species("A")
    my_mat = F.Material(D_0=[1, 1], E_D=0.1)

    with pytest.raises(ValueError, match="D_0 and E_D must be either floats or dicts"):
        my_mat.get_diffusion_coefficient(test_mesh.mesh, 500, species=A)


def test_error_raised_when_species_not_given_with_dict():
    """Test that a value error is raised when a species has not been given in
    the get_diffusion_coefficient function when using a dict for properties"""
    A = F.Species("A")
    B = F.Species("B")
    my_mat = F.Material(D_0={A: 1, B: 1}, E_D={A: 0.1, B: 0.1})

    with pytest.raises(
        ValueError, match="species must be provided if D_0 and E_D are dicts"
    ):
        my_mat.get_diffusion_coefficient(test_mesh.mesh, 500)


def test_error_raised_when_species_not_not_in_D_0_dict():
    """Test that a value error is raised when a species has not been given but
    has no value in the dict"""
    A = F.Species("A")
    B = F.Species("B")
    J = F.Species("J")
    my_mat = F.Material(D_0={A: 1, B: 1}, E_D={A: 0.1, B: 0.1})

    with pytest.raises(ValueError, match="J is not in D_0 keys"):
        my_mat.get_diffusion_coefficient(test_mesh.mesh, 500, species=J)


def test_D_0_raises_ValueError_if_species_not_provided_in_dict():
    """Test that a value error is raised in the get_diffusion_coefficient
    function"""
    # TODO remove this when material class is updated
    A = F.Species("A")
    B = F.Species("B")
    my_mat = F.Material(D_0={A: 1, B: 2}, E_D=1)

    with pytest.raises(ValueError, match="species must be provided if D_0 is a dict"):
        my_mat.get_D_0()


def test_D_0_raises_ValueError_if_species_given_not_in_dict_keys():
    """Test that a value error is raised in the get_diffusion_coefficient
    function"""
    # TODO remove this when material class is updated
    A = F.Species("A")
    B = F.Species("B")
    J = F.Species("J")
    my_mat = F.Material(D_0={A: 1, B: 2}, E_D=1)

    with pytest.raises(ValueError, match="J is not in D_0 keys"):
        my_mat.get_D_0(species=J)


def test_raises_TypeError_when_D_0_is_not_correct_type():
    """Test that a TypeError is raised when D_0 is not a float or a dict"""

    my_mat = F.Material(D_0=[1, 2], E_D=1)

    with pytest.raises(TypeError, match="D_0 must be either a float, int or a dict"):
        my_mat.get_D_0()


def test_E_D_raises_ValueError_if_species_not_provided_in_dict():
    """Test that a value error is raised in the get_diffusion_coefficient
    function"""
    # TODO remove this when material class is updated
    A = F.Species("A")
    B = F.Species("B")
    my_mat = F.Material(D_0=1, E_D={A: 1, B: 2})

    with pytest.raises(ValueError, match="species must be provided if E_D is a dict"):
        my_mat.get_E_D()


def test_E_D_raises_ValueError_if_species_given_not_in_dict_keys():
    """Test that a value error is raised in the get_diffusion_coefficient
    function"""
    # TODO remove this when material class is updated
    A = F.Species("A")
    B = F.Species("B")
    J = F.Species("J")
    my_mat = F.Material(D_0=1, E_D={A: 1, B: 2})

    with pytest.raises(ValueError, match="J is not in E_D keys"):
        my_mat.get_E_D(species=J)


def test_raises_TypeError_when_E_D_is_not_correct_type():
    """Test that a TypeError is raised when E_D is not a float or a dict"""

    my_mat = F.Material(D_0=1, E_D=[1, 2])

    with pytest.raises(TypeError, match="E_D must be either a float, int or a dict"):
        my_mat.get_E_D()



import numpy as np
import ufl
from dolfinx import fem


# TODO rename this to InitialConcentration and create a new base class
class InitialCondition:
    """
    Initial condition class

    Args:
        value (float, int, fem.Constant or callable): the value of the initial condition
        species (festim.Species): the species to which the condition is applied

    Attributes:
        value (float, int, fem.Constant or callable): the value of the initial condition
        species (festim.Species): the species to which the source is applied
        expr_fenics (LambdaType or fem.Expression): the value of the initial condition in
            fenics format

    Usage:
        >>> from festim import InitialCondition
        >>> InitialCondition(value=1, species=my_species)
        >>> InitialCondition(value=lambda x: 1 + x[0], species=my_species)
        >>> InitialCondition(value=lambda T: 1 + T, species=my_species)
        >>> InitialCondition(value=lambda x, T: 1 + x[0] + T, species=my_species)
    """

    def __init__(self, value, species):
        self.value = value
        self.species = species

        self.expr_fenics = None

    def create_expr_fenics(self, mesh, temperature, function_space):
        """Creates the expr_fenics of the initial condition.
        If the value is a float or int, a function is created with an array with
        the shape of the mesh and all set to the value.
        Otherwise, it is converted to a fem.Expression.

        Args:
            mesh (dolfinx.mesh.Mesh) : the mesh
            temperature (float): the temperature
            function_space(dolfinx.fem.FunctionSpaceBase): the function space of the species
        """
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.value, (int, float)):
            self.expr_fenics = lambda x: np.full(x.shape[1], self.value)

        elif callable(self.value):
            arguments = self.value.__code__.co_varnames
            kwargs = {}
            if "t" in arguments:
                raise ValueError("Initial condition cannot be a function of time.")
            if "x" in arguments:
                kwargs["x"] = x
            if "T" in arguments:
                kwargs["T"] = temperature

            self.expr_fenics = fem.Expression(
                self.value(**kwargs),
                function_space.element.interpolation_points(),
            )


class InitialTemperature:
    def __init__(self, value) -> None:
        self.value = value
        self.expr_fenics = None

    def create_expr_fenics(self, mesh, function_space):
        """Creates the expr_fenics of the initial condition.
        If the value is a float or int, a function is created with an array with
        the shape of the mesh and all set to the value.
        Otherwise, it is converted to a fem.Expression.

        Args:
            mesh (dolfinx.mesh.Mesh) : the mesh
            function_space(dolfinx.fem.FunctionSpace): the function space of the species
        """
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.value, (int, float)):
            self.expr_fenics = lambda x: np.full(x.shape[1], self.value)

        elif callable(self.value):
            arguments = self.value.__code__.co_varnames
            kwargs = {}
            if "t" in arguments:
                raise ValueError("Initial condition cannot be a function of time.")
            if "x" in arguments:
                kwargs["x"] = x

            self.expr_fenics = fem.Expression(
                self.value(**kwargs),
                function_space.element.interpolation_points(),
            )



from mpi4py import MPI

import basix.ufl
import dolfinx.mesh
import numpy as np
import ufl

import festim as F


class Mesh1D(F.Mesh):
    """
    1D Mesh

    Args:
        vertices (list or np.ndarray): the mesh x-coordinates (m)

    Attributes:
        vertices (np.ndarray): the mesh x-coordinates (m)
    """

    def __init__(self, vertices, **kwargs) -> None:
        self.vertices = vertices

        mesh = self.generate_mesh()
        super().__init__(mesh=mesh, **kwargs)

    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    def vertices(self, value):
        self._vertices = np.sort(np.unique(value)).astype(float)

    def generate_mesh(self):
        """Generates a 1D mesh"""
        degree = 1
        domain = ufl.Mesh(
            basix.ufl.element(basix.ElementFamily.P, "interval", degree, shape=(1,))
        )

        mesh_points = np.reshape(self.vertices, (len(self.vertices), 1))
        indexes = np.arange(self.vertices.shape[0])
        cells = np.stack((indexes[:-1], indexes[1:]), axis=-1)

        return dolfinx.mesh.create_mesh(MPI.COMM_WORLD, cells, mesh_points, domain)

    def check_borders(self, volume_subdomains):
        """Checks that the borders of the subdomain are within the domain

        Args:
            volume_subdomains (list of festim.VolumeSubdomain1D): the volume subdomains

        Raises:
            Value error: if borders outside the domain
        """
        # check that subdomains are connected
        all_borders = [border for vol in volume_subdomains for border in vol.borders]
        sorted_borders = np.sort(all_borders)
        for start, end in zip(sorted_borders[1:-2:2], sorted_borders[2:-1:2]):
            if start != end:
                raise ValueError("Subdomain borders don't match to each other")

        # check volume subdomain is defined
        # TODO this possible by default
        if len(all_borders) == 0:
            raise ValueError("No volume subdomains defined")

        # check that subdomains are within the domain
        if (
            sorted_borders[0] != self.vertices[0]
            or sorted_borders[-1] != self.vertices[-1]
        ):
            raise ValueError("borders dont match domain borders")

    def define_meshtags(self, surface_subdomains, volume_subdomains, interfaces=None):
        # check if all borders are defined
        self.check_borders(volume_subdomains)
        return super().define_meshtags(
            surface_subdomains, volume_subdomains, interfaces
        )



import dolfinx.mesh
import numpy as np
import numpy.typing as npt


class SurfaceSubdomain:
    """
    Surface subdomain class

    Args:
        id (int): the id of the surface subdomain
    """

    id: int

    def __init__(self, id):
        self.id = id

    def locate_boundary_facet_indices(
        self, mesh: dolfinx.mesh.Mesh
    ) -> npt.NDArray[np.int32]:
        """Locates the dof of the surface subdomain within the function space
        and return the index of the dof

        Args:
            mesh (dolfinx.mesh.Mesh): the mesh of the simulation

        Returns:
            index (np.array): the first value in the list of surface facet
                indices of the subdomain
        """
        fdim = mesh.topology.dim - 1
        mesh.topology.create_connectivity(fdim, fdim + 1)
        return dolfinx.mesh.exterior_facet_indices(mesh.topology)


def find_surface_from_id(id: int, surfaces: list):
    """Returns the correct surface subdomain object from a list of surface ids
    based on an int

    Args:
        id (int): the id of the surface subdomain
        surfaces (list of F.SurfaceSubdomain): the list of surfaces

    Returns:
        festim.SurfaceSubdomain: the surface subdomain object with the correct id

    Raises:
        ValueError: if the surface name is not found in the list of surfaces

    """
    for surf in surfaces:
        if surf.id == id:
            return surf
    raise ValueError(f"id {id} not found in list of surfaces")



import numpy as np
import pytest

import festim as F


@pytest.mark.parametrize("test_type", [int, F.Stepsize, float])
def test_stepsize_value(test_type):
    """Test that the stepsize is correctly set"""
    test_value = 23.0
    my_settings = F.Settings(atol=1, rtol=0.1)
    my_settings.stepsize = test_type(test_value)

    assert isinstance(my_settings.stepsize, F.Stepsize)
    assert np.isclose(my_settings.stepsize.initial_value, test_value)


def test_stepsize_value_wrong_type():
    """Checks that an error is raised when the wrong type is given"""
    my_settings = F.Settings(atol=1, rtol=0.1)

    with pytest.raises(TypeError):
        my_settings.stepsize = "coucou"



import festim as F
import numpy as np

import dolfinx.fem as fem
import ufl


class FluxFromSurfaceReaction(F.SurfaceFlux):
    def __init__(self, reaction: F.SurfaceReactionBC):
        super().__init__(
            F.Species(),  # just a dummy species here
            reaction.subdomain,
        )
        self.reaction = reaction.flux_bcs[0]

    def compute(self, ds):
        self.value = fem.assemble_scalar(
            fem.form(self.reaction.value_fenics * ds(self.surface.id))
        )
        self.data.append(self.value)


my_model = F.HydrogenTransportProblem()
my_model.mesh = F.Mesh1D(vertices=np.linspace(0, 1, 1000))
my_mat = F.Material(name="mat", D_0=1, E_D=0)
vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
left = F.SurfaceSubdomain1D(id=1, x=0)
right = F.SurfaceSubdomain1D(id=2, x=1)

my_model.subdomains = [vol, left, right]

H = F.Species("H")
D = F.Species("D")
my_model.species = [H, D]

my_model.temperature = 500

surface_reaction_hd = F.SurfaceReactionBC(
    reactant=[H, D],
    gas_pressure=0,
    k_r0=0.01,
    E_kr=0,
    k_d0=0,
    E_kd=0,
    subdomain=right,
)

surface_reaction_hh = F.SurfaceReactionBC(
    reactant=[H, H],
    gas_pressure=lambda t: ufl.conditional(ufl.gt(t, 1), 2, 0),
    k_r0=0.02,
    E_kr=0,
    k_d0=0.03,
    E_kd=0,
    subdomain=right,
)

surface_reaction_dd = F.SurfaceReactionBC(
    reactant=[D, D],
    gas_pressure=0,
    k_r0=0.01,
    E_kr=0,
    k_d0=0,
    E_kd=0,
    subdomain=right,
)

my_model.boundary_conditions = [
    F.DirichletBC(subdomain=left, value=2, species=H),
    F.DirichletBC(subdomain=left, value=2, species=D),
    surface_reaction_hd,
    surface_reaction_hh,
    surface_reaction_dd,
]

H_flux_right = F.SurfaceFlux(H, right)
H_flux_left = F.SurfaceFlux(H, left)
D_flux_right = F.SurfaceFlux(D, right)
D_flux_left = F.SurfaceFlux(D, left)
HD_flux = FluxFromSurfaceReaction(surface_reaction_hd)
HH_flux = FluxFromSurfaceReaction(surface_reaction_hh)
DD_flux = FluxFromSurfaceReaction(surface_reaction_dd)
my_model.exports = [
    F.XDMFExport("test.xdmf", H),
    H_flux_left,
    H_flux_right,
    D_flux_left,
    D_flux_right,
    HD_flux,
    HH_flux,
    DD_flux,
]

my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, final_time=5, transient=True)

my_model.settings.stepsize = 0.1

my_model.initialise()
my_model.run()

import matplotlib.pyplot as plt

plt.stackplot(
    H_flux_left.t,
    np.abs(H_flux_left.data),
    np.abs(D_flux_left.data),
    labels=["H_in", "D_in"],
)
plt.stackplot(
    H_flux_right.t,
    -np.abs(H_flux_right.data),
    -np.abs(D_flux_right.data),
    labels=["H_out", "D_out"],
)
plt.legend()
plt.xlabel("Time (s)")
plt.ylabel("Flux (atom/m^2/s)")
plt.figure()
plt.stackplot(
    HD_flux.t,
    np.abs(HH_flux.data),
    np.abs(HD_flux.data),
    np.abs(DD_flux.data),
    labels=["HH", "HD", "DD"],
)
plt.legend(reverse=True)
plt.xlabel("Time (s)")
plt.ylabel("Flux (molecule/m^2/s)")


plt.figure()
plt.plot(H_flux_right.t, -np.array(H_flux_right.data), label="from gradient (H)")
plt.plot(
    H_flux_right.t,
    2 * np.array(HH_flux.data) + np.array(HD_flux.data),
    linestyle="--",
    label="from reaction rates (H)",
)

plt.plot(D_flux_right.t, -np.array(D_flux_right.data), label="from gradient (D)")
plt.plot(
    D_flux_right.t,
    2 * np.array(DD_flux.data) + np.array(HD_flux.data),
    linestyle="--",
    label="from reaction rates (D)",
)
plt.xlabel("Time (s)")
plt.ylabel("Flux (atom/m^2/s)")
plt.legend()
plt.show()

# check that H_flux_right == 2*HH_flux + HD_flux
H_flux_from_gradient = -np.array(H_flux_right.data)
H_flux_from_reac = 2 * np.array(HH_flux.data) + np.array(HD_flux.data)
assert np.allclose(
    H_flux_from_gradient,
    H_flux_from_reac,
    rtol=0.5e-2,
    atol=0.005,
)
# check that D_flux_right == 2*DD_flux + HD_flux
D_flux_from_gradient = -np.array(D_flux_right.data)
D_flux_from_reac = 2 * np.array(DD_flux.data) + np.array(HD_flux.data)
assert np.allclose(
    D_flux_from_gradient,
    D_flux_from_reac,
    rtol=0.5e-2,
    atol=0.005,
)



import time

from mpi4py import MPI
from petsc4py import PETSc

import basix
import numpy as np
import tqdm.autonotebook
from dolfinx.fem import (
    Constant,
    Function,
    assemble_scalar,
    dirichletbc,
    form,
    functionspace,
    locate_dofs_geometrical,
    locate_dofs_topological,
)
from dolfinx.fem.petsc import (
    NonlinearProblem,
)
from dolfinx.io import XDMFFile
from dolfinx.mesh import create_mesh, locate_entities, meshtags
from dolfinx.nls.petsc import NewtonSolver
from test_permeation_problem import test_permeation_problem
from ufl import (
    FacetNormal,
    Measure,
    Mesh,
    TestFunction,
    dot,
    exp,
    grad,
)


def fenics_test_permeation_problem(mesh_size=1001):
    L = 3e-04
    indices = np.linspace(0, L, num=mesh_size)
    gdim, shape, degree = 1, "interval", 1
    domain = Mesh(basix.ufl.element("Lagrange", shape, degree, shape=(gdim,)))
    mesh_points = np.reshape(indices, (len(indices), 1))
    indexes = np.arange(mesh_points.shape[0])
    cells = np.stack((indexes[:-1], indexes[1:]), axis=-1)
    my_mesh = create_mesh(MPI.COMM_WORLD, cells, mesh_points, domain)
    fdim = my_mesh.topology.dim - 1
    vdim = my_mesh.topology.dim
    n = FacetNormal(my_mesh)

    elements = basix.ufl.element(
        basix.ElementFamily.P,
        my_mesh.basix_cell(),
        1,
        basix.LagrangeVariant.equispaced,
    )
    V = functionspace(my_mesh, elements)
    u = Function(V)
    u_n = Function(V)
    v = TestFunction(V)

    num_cells = my_mesh.topology.index_map(vdim).size_local
    cells = np.arange(num_cells, dtype=np.int32)
    markers = np.full(num_cells, 1, dtype=np.int32)
    markers[
        locate_entities(
            my_mesh,
            vdim,
            lambda x: np.logical_and(x[0] >= indices[0], x[0] <= indices[1]),
        )
    ] = 1
    mesh_tags_volumes = meshtags(my_mesh, vdim, cells, markers)

    dofs_L = locate_dofs_geometrical(V, lambda x: np.isclose(x[0], 0))
    dofs_R = locate_dofs_geometrical(V, lambda x: np.isclose(x[0], indices[-1]))

    dofs_facets = np.array([dofs_L[0], dofs_R[0]], dtype=np.int32)
    tags_facets = np.array([1, 2], dtype=np.int32)

    mesh_tags_facets = meshtags(my_mesh, fdim, dofs_facets, tags_facets)
    mesh_tags_volumes = meshtags(my_mesh, vdim, cells, markers)
    ds = Measure("ds", domain=my_mesh, subdomain_data=mesh_tags_facets)
    dx = Measure("dx", domain=my_mesh, subdomain_data=mesh_tags_volumes)

    temperature = 500
    k_B = 8.6173303e-5

    def siverts_law(T, S_0, E_S, pressure):
        S = S_0 * exp(-E_S / k_B / T)
        return S * pressure**0.5

    left_facets = mesh_tags_facets.find(1)
    my_mesh.topology.create_connectivity(fdim, my_mesh.topology.dim)
    left_dofs = locate_dofs_topological(V, fdim, left_facets)
    right_facets = mesh_tags_facets.find(2)
    right_dofs = locate_dofs_topological(V, fdim, right_facets)

    S_0 = 4.02e21
    E_S = 1.04
    P_up = 100
    surface_conc = siverts_law(T=temperature, S_0=S_0, E_S=E_S, pressure=P_up)
    bc_sieverts = dirichletbc(
        Constant(my_mesh, PETSc.ScalarType(surface_conc)), left_dofs, V
    )
    bc_outgas = dirichletbc(Constant(my_mesh, PETSc.ScalarType(0)), right_dofs, V)
    bcs = [bc_sieverts, bc_outgas]

    D_0 = 1.9e-7
    E_D = 0.2
    D = D_0 * exp(-E_D / k_B / temperature)

    dt = 1 / 20
    final_time = 50
    num_steps = int(final_time / dt)

    F = dot(D * grad(u), grad(v)) * dx(1)
    F += ((u - u_n) / dt) * v * dx(1)

    problem = NonlinearProblem(F, u, bcs=bcs)
    solver = NewtonSolver(MPI.COMM_WORLD, problem)
    solver.convergence_criterion = "incremental"
    solver.rtol = 1e-10
    solver.atol = 1e10
    solver.report = True
    ksp = solver.krylov_solver
    opts = PETSc.Options()
    option_prefix = ksp.getOptionsPrefix()
    opts[f"{option_prefix}ksp_type"] = "cg"
    opts[f"{option_prefix}pc_type"] = "gamg"
    opts[f"{option_prefix}pc_factor_mat_solver_type"] = "mumps"
    ksp.setFromOptions()

    mobile_xdmf = XDMFFile(MPI.COMM_WORLD, "mobile_concentration.xdmf", "w")
    mobile_xdmf.write_mesh(my_mesh)

    flux_values = []
    times = []
    t = 0
    progress = tqdm.autonotebook.tqdm(
        desc="Solving H transport problem", total=final_time, unit_scale=True
    )
    while t < final_time:
        progress.update(float(dt))
        t += float(dt)

        solver.solve(u)

        mobile_xdmf.write_function(u, t)

        surface_flux = form(D * dot(grad(u), n) * ds(2))
        flux = assemble_scalar(surface_flux)
        flux_values.append(flux)
        times.append(t)

        u_n.x.array[:] = u.x.array[:]

    mobile_xdmf.close()

    # analytical solution
    S = S_0 * exp(-E_S / k_B / float(temperature))
    permeability = float(D) * S
    times = np.array(times)

    n_array = np.arange(1, 10000)[:, np.newaxis]
    summation = np.sum(
        (-1) ** n_array * np.exp(-((np.pi * n_array) ** 2) * float(D) / L**2 * times),
        axis=0,
    )
    analytical_flux = P_up**0.5 * permeability / L * (2 * summation + 1)

    analytical_flux = np.abs(analytical_flux)
    flux_values = np.array(np.abs(flux_values))

    indices = np.where(analytical_flux > 0.01 * np.max(analytical_flux))
    analytical_flux = analytical_flux[indices]
    flux_values = flux_values[indices]

    relative_error = np.abs((flux_values - analytical_flux) / analytical_flux)

    error = relative_error.mean()

    assert error < 0.01


def test_festim_vs_fenics_permeation_benchmark():
    """Runs a problem with pure fenicsx and the same problem with FESTIM and
    raise ValueError if difference is too high"""

    start = time.time()
    fenics_test_permeation_problem(mesh_size=20001)
    fenics_time = time.time() - start

    start = time.time()
    test_permeation_problem(mesh_size=20001)
    festim_time = time.time() - start

    diff = (fenics_time - festim_time) / fenics_time
    threshold = -0.1
    if diff < threshold:
        raise ValueError(
            f"festim is {
                np.abs(diff):.1%
            } slower than fenics, current acceptable threshold of {
                np.abs(threshold):.1%
            }"
        )
    else:
        print(f"avg relative diff between festim and fenics {diff:.1%}")


if __name__ == "__main__":
    test_festim_vs_fenics_permeation_benchmark()



from mpi4py import MPI

import dolfinx.mesh
import numpy as np
import pytest
import ufl
from dolfinx import default_scalar_type, fem
from ufl.conditional import Conditional

import festim as F

dummy_mat = F.Material(D_0=1, E_D=1, name="dummy_mat")

mesh = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 10)


def test_init():
    """Test that the attributes are set correctly"""
    # create a DirichletBC object
    subdomain = F.SurfaceSubdomain1D(1, x=0)
    value = 1.0
    species = "test"
    bc = F.DirichletBC(subdomain, value, species)

    # check that the attributes are set correctly
    assert bc.subdomain == subdomain
    assert bc.value == value
    assert bc.species == species
    assert bc.value_fenics is None
    assert bc.bc_expr is None


def test_value_fenics():
    """Test that the value_fenics attribute can be set to a valid value
    and that an invalid type throws an error
    """
    # create a DirichletBC object
    subdomain = F.SurfaceSubdomain1D(1, x=0)
    value = 1.0
    species = "test"
    bc = F.DirichletBC(subdomain, value, species)

    # set the value_fenics attribute to a valid value
    value_fenics = fem.Constant(mesh, 2.0)
    bc.value_fenics = value_fenics

    # check that the value_fenics attribute is set correctly
    assert bc.value_fenics == value_fenics

    # set the value_fenics attribute to an invalid value
    with pytest.raises(TypeError):
        bc.value_fenics = "invalid"


def test_callable_for_value():
    """Test that the value attribute can be a callable function of x and t"""

    subdomain = F.SurfaceSubdomain1D(1, x=1)
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)

    def value(x, t):
        return 1.0 + x[0] + t

    species = F.Species("test")

    bc = F.DirichletBC(subdomain, value, species)

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh(mesh),
        subdomains=[subdomain, vol_subdomain],
        species=[species],
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()

    T = fem.Constant(my_model.mesh.mesh, 550.0)
    t = fem.Constant(my_model.mesh.mesh, 0.0)
    bc.create_value(my_model.function_space, T, t)

    # check that the value_fenics attribute is set correctly
    assert isinstance(bc.value_fenics, fem.Function)

    # check the initial value of the boundary condition
    assert bc.value_fenics.x.petsc_vec.array[-1] == float(
        value(x=np.array([subdomain.x]), t=0.0)
    )

    # check the value of the boundary condition after updating the time
    for i in range(10):
        t.value = i
        bc.update(float(t))
        expected_value = float(value(x=np.array([subdomain.x]), t=float(t)))
        computed_value = bc.value_fenics.x.petsc_vec.array[-1]
        assert np.isclose(computed_value, expected_value)


def test_value_callable_x_t_T():
    """Test that the value attribute can be a callable function of x, t, and T"""

    subdomain = F.SurfaceSubdomain1D(1, x=1)
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)

    def value(x, t, T):
        return 1.0 + x[0] + t + T

    species = F.Species("test")

    bc = F.DirichletBC(subdomain, value, species)

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh(mesh),
        subdomains=[subdomain, vol_subdomain],
        species=[species],
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()

    T = fem.Constant(my_model.mesh.mesh, 550.0)
    t = fem.Constant(my_model.mesh.mesh, 0.0)
    bc.create_value(my_model.function_space, T, t)

    # check that the value_fenics attribute is set correctly
    assert isinstance(bc.value_fenics, fem.Function)

    # check the initial value of the boundary condition
    assert np.isclose(
        bc.value_fenics.x.petsc_vec.array[-1],
        float(value(x=np.array([subdomain.x]), t=float(t), T=float(T))),
    )

    # check the value of the boundary condition after updating the time
    for i in range(10):
        t.value = i
        T.value += i
        bc.update(float(t))

        expected_value = float(value(x=np.array([subdomain.x]), t=float(t), T=float(T)))
        computed_value = bc.value_fenics.x.petsc_vec.array[-1]
        assert np.isclose(computed_value, expected_value)


@pytest.mark.parametrize("value", [lambda t: t, lambda t: 1.0 + t])
def test_callable_t_only(value):
    """Test that the value attribute can be a callable function of t only"""

    subdomain = F.SurfaceSubdomain1D(1, x=1)
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)
    species = F.Species("test")

    bc = F.DirichletBC(subdomain, value, species)

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh(mesh),
        subdomains=[subdomain, vol_subdomain],
        species=[species],
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()

    T = fem.Constant(my_model.mesh.mesh, 550.0)
    t = fem.Constant(my_model.mesh.mesh, 0.0)
    bc.create_value(my_model.function_space, T, t)

    # check that the value_fenics attribute is set correctly
    assert isinstance(bc.value_fenics, fem.Constant)

    # check the initial value of the boundary condition
    assert np.isclose(
        float(bc.value_fenics),
        float(value(t=float(t))),
    )

    # check the value of the boundary condition after updating the time
    for i in range(10):
        t.value = i
        bc.update(float(t))

        expected_value = float(value(t=float(t)))
        computed_value = float(bc.value_fenics)
        assert np.isclose(computed_value, expected_value)


def test_callable_x_only():
    """Test that the value attribute can be a callable function of x only"""

    # BUILD
    subdomain = F.SurfaceSubdomain1D(1, x=1)
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)

    def value(x):
        return 1.0 + x[0]

    species = F.Species("test")

    bc = F.DirichletBC(subdomain, value, species)

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh(mesh),
        subdomains=[subdomain, vol_subdomain],
        species=[species],
    )

    my_model.define_function_spaces()
    my_model.define_meshtags_and_measures()

    T = fem.Constant(my_model.mesh.mesh, 550.0)
    t = fem.Constant(my_model.mesh.mesh, 0.0)

    # TEST
    bc.create_value(my_model.function_space, T, t)

    # check that the value_fenics attribute is set correctly
    assert isinstance(bc.value_fenics, fem.Function)

    # check the initial value of the boundary condition
    assert np.isclose(
        bc.value_fenics.x.petsc_vec.array[-1],
        float(value(x=np.array([subdomain.x]))),
    )

    # check the value of the boundary condition after updating the time
    for i in range(10):
        t.value = i
        bc.update(float(t))
        expected_value = float(value(x=np.array([subdomain.x])))
        computed_value = bc.value_fenics.x.petsc_vec.array[-1]
        assert np.isclose(computed_value, expected_value)


@pytest.mark.parametrize(
    "value",
    [
        1.0,
        lambda t: t,
        lambda t: 1.0 + t,
        lambda x: 1.0 + x[0],
        lambda x, t: 1.0 + x[0] + t,
        lambda x, t, T: 1.0 + x[0] + t + T,
        lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0),
        lambda t: 100.0 if t < 1 else 0.0,
    ],
)
def test_integration_with_HTransportProblem(value):
    """test that different callable functions can be applied to a dirichlet
    boundary condition, asserting in each case they match an expected value"""
    subdomain = F.SurfaceSubdomain1D(1, x=1)
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh(mesh),
        subdomains=[vol_subdomain, subdomain],
    )
    my_model.species = [F.Species("H")]
    my_bc = F.DirichletBC(subdomain, value, "H")
    my_model.boundary_conditions = [my_bc]

    my_model.temperature = fem.Constant(my_model.mesh.mesh, 550.0)

    my_model.settings = F.Settings(atol=1, rtol=0.1, final_time=2)
    my_model.settings.stepsize = F.Stepsize(initial_value=1)

    # RUN

    my_model.initialise()

    assert my_bc.value_fenics is not None

    my_model.run()

    # TEST

    if isinstance(value, float):
        expected_value = value
        computed_value = float(my_bc.value_fenics)
    elif callable(value):
        arguments = value.__code__.co_varnames
        if "x" in arguments and "t" in arguments and "T" in arguments:
            expected_value = value(x=np.array([subdomain.x]), t=2.0, T=550.0)
            computed_value = my_bc.value_fenics.x.petsc_vec.array[-1]
        elif "x" in arguments and "t" in arguments:
            expected_value = value(x=np.array([subdomain.x]), t=2.0)
            computed_value = my_bc.value_fenics.x.petsc_vec.array[-1]
        elif "x" in arguments:
            expected_value = value(x=np.array([subdomain.x]))
            computed_value = my_bc.value_fenics.x.petsc_vec.array[-1]
        elif "t" in arguments:
            expected_value = value(t=2.0)
            computed_value = float(my_bc.value_fenics)
        else:
            # test fails if lambda function is not recognised
            raise ValueError("value function not recognised")

    if isinstance(expected_value, Conditional):
        expected_value = float(expected_value)
    assert np.isclose(computed_value, expected_value)


@pytest.mark.parametrize(
    "value",
    [
        lambda t: ufl.conditional(ufl.lt(t, 1.0), 1, 2),
        lambda t: 1 + ufl.conditional(ufl.lt(t, 1.0), 1, 2.0),
        lambda t: 2 * ufl.conditional(ufl.lt(t, 1.0), 1, 2.0),
        lambda t: 2 / ufl.conditional(ufl.lt(t, 1.0), 1, 2.0),
    ],
)
def test_define_value_error_if_ufl_conditional_t_only(value):
    """Test that a ValueError is raised when the value attribute is a callable
    of t only and contains a ufl conditional"""

    subdomain = F.SurfaceSubdomain1D(1, x=1)
    species = F.Species("test")

    bc = F.DirichletBC(subdomain, value, species)

    t = fem.Constant(mesh, 0.0)
    V = dolfinx.fem.functionspace(mesh, ("Lagrange", 1))
    with pytest.raises(
        ValueError, match="self.value should return a float or an int, not "
    ):
        bc.create_value(V, temperature=None, t=t)


def test_species_predefined():
    """Test a ValueError is raised when the species defined in the boundary
    condition is not predefined in the model"""

    subdomain = F.SurfaceSubdomain1D(1, x=1)
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh(mesh),
        subdomains=[vol_subdomain, subdomain],
    )
    my_model.species = [F.Species("H")]
    my_bc = F.DirichletBC(subdomain, 1.0, "J")
    my_model.boundary_conditions = [my_bc]
    my_model.temperature = 1
    my_model.settings = F.Settings(atol=1, rtol=0.1)
    my_model.settings.stepsize = 1

    with pytest.raises(ValueError):
        my_model.initialise()


@pytest.mark.parametrize(
    "value_A, value_B",
    [
        (1.0, 1.0),
        (1.0, lambda t: t),
        (1.0, lambda t: 1.0 + t),
        (1.0, lambda x: 1.0 + x[0]),
        (1.0, lambda x, t: 1.0 + x[0] + t),
        (1.0, lambda x, t, T: 1.0 + x[0] + t + T),
        (1.0, lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0)),
    ],
)
def test_integration_with_a_multispecies_HTransportProblem(value_A, value_B):
    """test that a mixture of callable functions can be applied to dirichlet
    boundary conditions in a multispecies case, asserting in each case they
    match an expected value"""
    subdomain_A = F.SurfaceSubdomain1D(1, x=0)
    subdomain_B = F.SurfaceSubdomain1D(2, x=1)
    vol_subdomain = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)

    my_model = F.HydrogenTransportProblem(
        mesh=F.Mesh(mesh),
        subdomains=[vol_subdomain, subdomain_A, subdomain_B],
    )
    my_model.species = [F.Species("A"), F.Species("B")]
    my_bc_A = F.DirichletBC(subdomain_A, value_A, "A")
    my_bc_B = F.DirichletBC(subdomain_B, value_B, "B")
    my_model.boundary_conditions = [my_bc_A, my_bc_B]

    my_model.temperature = fem.Constant(my_model.mesh.mesh, 550.0)

    my_model.settings = F.Settings(atol=1, rtol=0.1, final_time=2)
    my_model.settings.stepsize = F.Stepsize(initial_value=1)

    # RUN

    my_model.initialise()

    for bc in [my_bc_A, my_bc_B]:
        assert bc.value_fenics is not None

    my_model.run()

    # TEST

    expected_value = value_A
    computed_value = float(my_bc_A.value_fenics)

    if isinstance(value_B, float):
        expected_value = value_B
        computed_value = float(my_bc_B.value_fenics)
    elif callable(value_B):
        arguments = value_B.__code__.co_varnames
        if "x" in arguments and "t" in arguments and "T" in arguments:
            expected_value = value_B(x=np.array([subdomain_B.x]), t=2.0, T=550.0)
            computed_value = my_bc_B.value_fenics.x.petsc_vec.array[-1]
        elif "x" in arguments and "t" in arguments:
            expected_value = value_B(x=np.array([subdomain_B.x]), t=2.0)
            computed_value = my_bc_B.value_fenics.x.petsc_vec.array[-1]
        elif "x" in arguments:
            expected_value = value_B(x=np.array([subdomain_B.x]))
            computed_value = my_bc_B.value_fenics.x.petsc_vec.array[-1]
        elif "t" in arguments:
            expected_value = value_B(t=2.0)
            computed_value = float(my_bc_B.value_fenics)
        else:
            # test fails if lambda function is not recognised
            raise ValueError("value function not recognised")

    if isinstance(expected_value, Conditional):
        expected_value = float(expected_value)
    assert np.isclose(computed_value, expected_value)


@pytest.mark.parametrize(
    "input, expected_value",
    [
        (1.0, False),
        (None, False),
        (fem.Constant(mesh, default_scalar_type(1.0)), False),
        (lambda t: t, True),
        (lambda t: 1.0 + t, True),
        (lambda x: 1.0 + x[0], False),
        (lambda x, t: 1.0 + x[0] + t, True),
        (lambda x, t, T: 1.0 + x[0] + t + T, True),
        (lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0), True),
    ],
)
def test_bc_time_dependent_attribute(input, expected_value):
    """Test that the time_dependent attribute is correctly set"""
    subdomain = F.SurfaceSubdomain1D(1, x=0)
    species = F.Species("test")
    my_bc = F.DirichletBC(subdomain, input, species)

    assert my_bc.time_dependent is expected_value


@pytest.mark.parametrize(
    "input, expected_value",
    [
        (1.0, False),
        (None, False),
        (fem.Constant(mesh, default_scalar_type(1.0)), False),
        (lambda T: T, True),
        (lambda t: 1.0 + t, False),
        (lambda x, T: 1.0 + x[0] + T, True),
        (lambda x, t, T: 1.0 + x[0] + t + T, True),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0),
            False,
        ),
    ],
)
def test_bc_temperature_dependent_attribute(input, expected_value):
    """Test that the temperature_dependent attribute is correctly set"""
    subdomain = F.SurfaceSubdomain1D(1, x=0)
    species = F.Species("test")
    my_bc = F.DirichletBC(subdomain, input, species)

    assert my_bc.temperature_dependent is expected_value



import dolfinx.mesh
import numpy as np

from .surface_subdomain import SurfaceSubdomain


class SurfaceSubdomain1D(SurfaceSubdomain):
    """
    Surface subdomain class for 1D cases

    Args:
        id (int): the id of the surface subdomain
        x (float): the x coordinate of the surface subdomain

    Attributes:
        id (int): the id of the surface subdomain
        x (float): the x coordinate of the surface subdomain

    Usage:
        >>> surf_subdomain = F.SurfaceSubdomain1D(id=1, x=1)
    """

    # FIXME: Rename this to _id and use getter/setter
    id: int
    x: float

    def __init__(self, id: int, x: float) -> None:
        super().__init__(id)
        self.x = x

    def locate_boundary_facet_indices(self, mesh: dolfinx.mesh.Mesh):
        """Locates the dof of the surface subdomain within the function space
        and return the index of the dof

        Args:
            mesh (dolfinx.mesh.Mesh): the mesh of the simulation

        Returns:
            index (np.array): the first value in the list of surface facet
                indices of the subdomain
        """
        assert mesh.geometry.dim == 1, "This method is only for 1D meshes"
        indices = dolfinx.mesh.locate_entities_boundary(
            mesh, 0, lambda x: np.isclose(x[0], self.x)
        )
        return indices



__all__ = ["VolumeSubdomain", "SurfaceSubdomain", "Interface", "Subdomain"]


from .interface import Interface
from .surface_subdomain import SurfaceSubdomain
from .volume_subdomain import VolumeSubdomain



from importlib import metadata

try:
    __version__ = metadata.version("FESTIM")
except Exception:
    __version__ = "unknown"


R = 8.314462618  # Gas constant J.mol-1.K-1
k_B = 8.6173303e-5  # Boltzmann constant eV.K-1


from .boundary_conditions.dirichlet_bc import (
    DirichletBC,
    DirichletBCBase,
    FixedConcentrationBC,
    FixedTemperatureBC,
)
from .boundary_conditions.flux_bc import FluxBCBase, HeatFluxBC, ParticleFluxBC
from .boundary_conditions.henrys_bc import HenrysBC
from .boundary_conditions.flux_bc import FluxBCBase, ParticleFluxBC, HeatFluxBC
from .boundary_conditions.surface_reaction import SurfaceReactionBC

from .boundary_conditions.sieverts_bc import SievertsBC
from .exports.average_surface import AverageSurface
from .exports.average_volume import AverageVolume
from .exports.maximum_surface import MaximumSurface
from .exports.maximum_volume import MaximumVolume
from .exports.minimum_surface import MinimumSurface
from .exports.minimum_volume import MinimumVolume
from .exports.surface_flux import SurfaceFlux
from .exports.surface_quantity import SurfaceQuantity
from .exports.total_surface import TotalSurface
from .exports.total_volume import TotalVolume
from .exports.volume_quantity import VolumeQuantity
from .exports.vtx import VTXSpeciesExport, VTXTemperatureExport
from .exports.xdmf import XDMFExport
from .heat_transfer_problem import HeatTransferProblem
from .helpers import as_fenics_constant
from .hydrogen_transport_problem import (
    HTransportProblemDiscontinuous,
    HydrogenTransportProblem,
    HTransportProblemPenalty,
)
from .problem_change_of_var import HydrogenTransportProblemDiscontinuousChangeVar
from .initial_condition import InitialCondition, InitialTemperature
from .material import Material
from .mesh.mesh import Mesh
from .mesh.mesh_1d import Mesh1D
from .mesh.mesh_from_xdmf import MeshFromXDMF
from .problem import ProblemBase
from .reaction import Reaction
from .settings import Settings
from .source import HeatSource, ParticleSource, SourceBase
from .species import ImplicitSpecies, Species, find_species_from_name, SpeciesChangeVar
from .stepsize import Stepsize
from .subdomain.interface import Interface
from .subdomain.surface_subdomain import SurfaceSubdomain, find_surface_from_id
from .subdomain.surface_subdomain_1d import SurfaceSubdomain1D
from .subdomain.volume_subdomain import VolumeSubdomain, find_volume_from_id
from .subdomain.volume_subdomain_1d import VolumeSubdomain1D
from .trap import Trap



from mpi4py import MPI

import dolfinx.mesh
import numpy as np
import pytest
import ufl
import ufl.core
from dolfinx import default_scalar_type, fem

import festim as F

dummy_mat = F.Material(D_0=1, E_D=1, name="dummy_mat")
mesh = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 10)


def test_init():
    """Test that the attributes are set correctly"""
    # create a DirichletBC object
    subdomain = F.SurfaceSubdomain1D(1, x=0)
    value = 1.0
    species = "test"
    bc = F.ParticleFluxBC(subdomain, value, species)

    # check that the attributes are set correctly
    assert bc.subdomain == subdomain
    assert bc.value == value
    assert bc.species == species
    assert bc.value_fenics is None
    assert bc.bc_expr is None


@pytest.mark.parametrize(
    "value, expected_type",
    [
        (1.0, fem.Constant),
        (lambda t: t, fem.Constant),
        (lambda t: 1.0 + t, fem.Constant),
        (lambda x: 1.0 + x[0], ufl.core.expr.Expr),
        (lambda x, t: 1.0 + x[0] + t, ufl.core.expr.Expr),
        (lambda x, t, T: 1.0 + x[0] + t + T, ufl.core.expr.Expr),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0),
            ufl.core.expr.Expr,
        ),
        (lambda t: 100.0 if t < 1 else 0.0, fem.Constant),
    ],
)
def test_create_value_fenics_type(value, expected_type):
    """Test that"""
    # BUILD
    left = F.SurfaceSubdomain1D(1, x=0)
    my_species = F.Species("test")
    T = F.as_fenics_constant(1, mesh)
    t = F.as_fenics_constant(0, mesh)
    bc = F.ParticleFluxBC(subdomain=left, value=value, species=my_species)

    # RUN
    bc.create_value_fenics(mesh, T, t)

    # TEST
    # check that the value_fenics attribute is set correctly
    assert isinstance(bc.value_fenics, expected_type)


@pytest.mark.parametrize(
    "value, expected_value",
    [
        (1.0, 1.0),
        (lambda t: t, 0.0),
        (lambda t: 4.0 + t, 4.0),
        (lambda t: 50.0 if t < 1 else 0.0, 50),
    ],
)
def test_create_value_fenics_value(value, expected_value):
    """Test that"""
    # BUILD
    left = F.SurfaceSubdomain1D(1, x=0)
    my_species = F.Species("test")
    T = F.as_fenics_constant(1, mesh)
    t = F.as_fenics_constant(0, mesh)
    bc = F.ParticleFluxBC(subdomain=left, value=value, species=my_species)

    # RUN
    bc.create_value_fenics(mesh, T, t)

    # TEST
    # check that the value_fenics attribute is set correctly
    if isinstance(bc.value_fenics, fem.Constant):
        assert np.isclose(bc.value_fenics.value, expected_value)


def test_create_value_fenics_dependent_conc():
    """Test that the value_fenics of ParticleFluxBC is set correctly when the value is dependent on the concentration"""
    # BUILD
    left = F.SurfaceSubdomain1D(1, x=0)
    my_species = F.Species("test")
    my_species.solution = F.as_fenics_constant(12, mesh)
    T = F.as_fenics_constant(1, mesh)
    t = F.as_fenics_constant(0, mesh)
    bc = F.ParticleFluxBC(
        subdomain=left,
        value=lambda c: 1.0 + c,
        species=my_species,
        species_dependent_value={"c": my_species},
    )

    # RUN
    bc.create_value_fenics(mesh, T, t)

    # TEST
    assert isinstance(bc.value_fenics, ufl.core.expr.Expr)
    assert bc.value_fenics == 1.0 + my_species.solution


def test_value_fenics_setter_error():
    left = F.SurfaceSubdomain1D(1, x=0)
    my_species = F.Species("test")
    bc = F.ParticleFluxBC(subdomain=left, value=1.0, species=my_species)

    with pytest.raises(
        TypeError,
        match="Value must be a dolfinx.fem.Function, dolfinx.fem.Constant, np.ndarray or ufl.core.expr.Expr not <class 'str'>",
    ):
        bc.value_fenics = "coucou"


def test_ValueError_raised_when_callable_returns_wrong_type():
    """The create_value_fenics method should raise a ValueError when the callable
    returns an object which is not a float or int"""

    surface = F.SurfaceSubdomain(id=1)
    species = F.Species("test")

    def my_value(t):
        return ufl.conditional(ufl.lt(t, 0.5), 100, 0)

    bc = F.ParticleFluxBC(subdomain=surface, value=my_value, species=species)

    T = fem.Constant(mesh, 550.0)
    t = fem.Constant(mesh, 0.0)

    with pytest.raises(
        ValueError,
        match="self.value should return a float or an int, not <class 'ufl.conditional.Conditional'",
    ):
        bc.create_value_fenics(mesh, T, t)


@pytest.mark.parametrize(
    "input, expected_value",
    [
        (1.0, False),
        (fem.Constant(mesh, default_scalar_type(1.0)), False),
        (lambda t: t, True),
        (lambda t: 1.0 + t, True),
        (lambda x: 1.0 + x[0], False),
        (lambda x, t: 1.0 + x[0] + t, True),
        (lambda x, t, T: 1.0 + x[0] + t + T, True),
        (lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0), True),
    ],
)
def test_bc_time_dependent_attribute(input, expected_value):
    """Test that the time_dependent attribute is correctly set"""
    surface = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)
    my_species = F.Species("test")
    my_bc = F.ParticleFluxBC(subdomain=surface, value=input, species=my_species)

    assert my_bc.time_dependent is expected_value


def test_bc_time_dependent_attribute_raises_error_when_value_none():
    """Test that the time_dependent attribute raises a TypeError when the value is None"""
    surface = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)
    my_flux_bc = F.FluxBCBase(subdomain=surface, value=None)

    with pytest.raises(
        TypeError,
        match="Value must be given to determine if its time dependent",
    ):
        my_flux_bc.time_dependent


@pytest.mark.parametrize(
    "input, expected_value",
    [
        (1.0, False),
        (None, False),
        (fem.Constant(mesh, default_scalar_type(1.0)), False),
        (lambda T: T, True),
        (lambda t: 1.0 + t, False),
        (lambda x, T: 1.0 + x[0] + T, True),
        (lambda x, t, T: 1.0 + x[0] + t + T, True),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0),
            False,
        ),
    ],
)
def test_bc_temperature_dependent_attribute(input, expected_value):
    """Test that the temperature_dependent attribute is correctly set"""
    surface = F.VolumeSubdomain1D(1, borders=[0, 1], material=dummy_mat)
    my_species = F.Species("test")
    my_bc = F.ParticleFluxBC(subdomain=surface, value=input, species=my_species)

    assert my_bc.temperature_dependent is expected_value


def test_HeatFluxBC_init():
    """Test that the attributes are set correctly"""
    # create a DirichletBC object
    subdomain = F.SurfaceSubdomain1D(1, x=0)
    value = 1.0
    bc = F.HeatFluxBC(subdomain, value)

    # check that the attributes are set correctly
    assert bc.subdomain == subdomain
    assert bc.value == value
    assert bc.value_fenics is None
    assert bc.bc_expr is None


@pytest.mark.parametrize(
    "value, expected_type",
    [
        (1.0, fem.Constant),
        (lambda t: t, fem.Constant),
        (lambda t: 1.0 + t, fem.Constant),
        (lambda x: 1.0 + x[0], ufl.core.expr.Expr),
        (lambda x, t: 1.0 + x[0] + t, ufl.core.expr.Expr),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 100.0 + x[0], 0.0),
            ufl.core.expr.Expr,
        ),
        (lambda t: 100.0 if t < 1 else 0.0, fem.Constant),
    ],
)
def test_create_value_fenics_type_HeatFluxBC(value, expected_type):
    """Test that"""
    # BUILD
    left = F.SurfaceSubdomain1D(1, x=0)
    t = F.as_fenics_constant(0, mesh)
    bc = F.HeatFluxBC(subdomain=left, value=value)
    temperature = F.as_fenics_constant(1, mesh)

    # RUN
    bc.create_value_fenics(mesh, temperature, t)

    # TEST
    # check that the value_fenics attribute is set correctly
    assert isinstance(bc.value_fenics, expected_type)


@pytest.mark.parametrize(
    "value, expected_value",
    [
        (1.0, 1.0),
        (lambda t: t, 0.0),
        (lambda t: 4.0 + t, 4.0),
        (lambda x: 1.0 + x[0], 2.0),
        (lambda x, t: 3.0 + x[0] + t, 4.0),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.0), 50.0 + x[0], 0.0),
            51,
        ),
        (lambda t: 50.0 if t < 1 else 0.0, 50),
    ],
)
def test_create_value_fenics_value_HeatFluxBC(value, expected_value):
    """Test that"""
    # BUILD
    left = F.SurfaceSubdomain1D(1, x=0)
    t = F.as_fenics_constant(0, mesh)
    bc = F.HeatFluxBC(subdomain=left, value=value)
    temperature = F.as_fenics_constant(1, mesh)

    # RUN
    bc.create_value_fenics(mesh, temperature, t)

    # TEST
    # check that the value_fenics attribute is set correctly
    if isinstance(bc.value_fenics, fem.Constant):
        assert np.isclose(bc.value_fenics.value, expected_value)

    if isinstance(bc.value_fenics, fem.Function):
        assert np.isclose(bc.value_fenics.x.array[-1], expected_value)



import os

import mpi4py.MPI as MPI

import dolfinx
import numpy as np
import pytest
import tqdm.autonotebook
import ufl
from dolfinx import fem
from dolfinx.io import XDMFFile

import festim as F


def source_from_exact_solution(
    exact_solution, thermal_conductivity, density, heat_capacity
):
    import sympy as sp
    from sympy.vector import CoordSys3D, divergence, gradient

    R = CoordSys3D("R")
    x = [R.x, R.y, R.z]
    t = sp.symbols("t")

    u = exact_solution(x, t)
    density = density(x, t)
    heat_capacity = heat_capacity(x, t)
    thermal_cond = thermal_conductivity(x, t)
    source = density * heat_capacity * sp.diff(u, t) - divergence(
        thermal_cond * gradient(u, x), x
    )

    return sp.lambdify([x, t], source)


def error_L2(u_computed, u_exact, degree_raise=3):
    # Create higher order function space
    degree = u_computed.function_space.ufl_element().degree
    family = u_computed.function_space.ufl_element().family_name
    mesh = u_computed.function_space.mesh
    W = fem.functionspace(mesh, (family, degree + degree_raise))
    # Interpolate approximate solution
    u_W = fem.Function(W)
    u_W.interpolate(u_computed)

    # Interpolate exact solution, special handling if exact solution
    # is a ufl expression or a python lambda function
    u_ex_W = fem.Function(W)
    if isinstance(u_exact, ufl.core.expr.Expr):
        u_expr = fem.Expression(u_exact, W.element.interpolation_points)
        u_ex_W.interpolate(u_expr)
    else:
        u_ex_W.interpolate(u_exact)

    # Compute the error in the higher order function space
    e_W = fem.Function(W)
    e_W.x.array[:] = u_W.x.array - u_ex_W.x.array

    # Integrate the error
    error = fem.form(ufl.inner(e_W, e_W) * ufl.dx)
    error_local = fem.assemble_scalar(error)
    error_global = mesh.comm.allreduce(error_local, op=MPI.SUM)
    return np.sqrt(error_global)


def test_MMS_1():
    thermal_conductivity = 4.0

    def exact_solution(x):
        return 2 * x[0] ** 2

    mms_source = -4 * thermal_conductivity

    my_problem = F.HeatTransferProblem()

    my_problem.mesh = F.Mesh1D(vertices=np.linspace(0, 1, 2000))
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)
    mat = F.Material(D_0=None, E_D=None)
    mat.thermal_conductivity = thermal_conductivity

    my_problem.subdomains = [
        left,
        right,
        F.VolumeSubdomain1D(id=1, borders=[0, 1], material=mat),
    ]

    my_problem.boundary_conditions = [
        F.FixedTemperatureBC(subdomain=left, value=exact_solution),
        F.FixedTemperatureBC(subdomain=right, value=exact_solution),
    ]

    my_problem.sources = [
        F.HeatSource(value=mms_source, volume=my_problem.volume_subdomains[0])
    ]

    my_problem.settings = F.Settings(
        atol=1e-10,
        rtol=1e-10,
        transient=False,
    )

    my_problem.initialise()
    my_problem.run()

    computed_solution = my_problem.u
    L2_error = error_L2(computed_solution, exact_solution)
    assert L2_error < 1e-7


def test_MMS_T_dependent_thermal_cond():
    """MMS test with space T dependent thermal cond"""

    def thermal_conductivity(T):
        return 3 * T + 2

    def exact_solution(x):
        return 2 * x[0] ** 2 + 1

    # TODO would be nice to automate
    def mms_source(x):
        return -(72 * x[0] ** 2 + 20)

    my_problem = F.HeatTransferProblem()

    my_problem.mesh = F.Mesh1D(vertices=np.linspace(0, 1, 2000))
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)
    mat = F.Material(D_0=None, E_D=None)
    mat.thermal_conductivity = thermal_conductivity

    my_problem.subdomains = [
        left,
        right,
        F.VolumeSubdomain1D(id=1, borders=[0, 1], material=mat),
    ]
    my_problem.boundary_conditions = [
        F.FixedTemperatureBC(subdomain=left, value=exact_solution),
        F.FixedTemperatureBC(subdomain=right, value=exact_solution),
    ]

    my_problem.sources = [
        F.HeatSource(value=mms_source, volume=my_problem.volume_subdomains[0])
    ]

    my_problem.settings = F.Settings(
        atol=1e-10,
        rtol=1e-10,
        transient=False,
    )

    my_problem.initialise()
    my_problem.run()

    computed_solution = my_problem.u
    L2_error = error_L2(computed_solution, exact_solution)
    assert L2_error < 1e-7


def test_heat_transfer_transient():
    """
    MMS test for transient heat transfer
    constant thermal conductivity density and heat capacity
    """
    density = 2
    heat_capacity = 3
    thermal_conductivity = 4

    def exact_solution(x, t):
        return 2 * x[0] ** 2 + 20 * t

    dTdt = 20
    mms_source = density * heat_capacity * dTdt - thermal_conductivity * 4

    my_problem = F.HeatTransferProblem()

    my_problem.mesh = F.Mesh1D(vertices=np.linspace(2, 3, 2000))
    left = F.SurfaceSubdomain1D(id=1, x=2)
    right = F.SurfaceSubdomain1D(id=2, x=3)
    mat = F.Material(D_0=None, E_D=None)
    mat.thermal_conductivity = thermal_conductivity
    mat.density = density
    mat.heat_capacity = heat_capacity

    my_problem.subdomains = [
        left,
        right,
        F.VolumeSubdomain1D(id=1, borders=[2, 3], material=mat),
    ]
    # NOTE: it's good to check that without the IC the solution is not the exact one
    my_problem.initial_condition = F.InitialTemperature(lambda x: exact_solution(x, 0))

    my_problem.boundary_conditions = [
        F.FixedTemperatureBC(subdomain=left, value=exact_solution),
        F.FixedTemperatureBC(subdomain=right, value=exact_solution),
    ]

    my_problem.sources = [
        F.HeatSource(value=mms_source, volume=my_problem.volume_subdomains[0])
    ]

    my_problem.settings = F.Settings(
        atol=1e-8,
        rtol=1e-10,
        final_time=1,  # final time shouldn't be too long so that a potential error at the initial timestep is not negligible
    )

    # Forward euler isn't great so dt should be small
    # although it's ok here since the time derivative is constant
    my_problem.settings.stepsize = F.Stepsize(0.1)

    my_problem.exports = [
        F.VTXTemperatureExport(filename="test_transient_heat_transfer.bp")
    ]

    my_problem.initialise()
    my_problem.run()

    computed_solution = my_problem.u
    # we use the exact final time of the simulation which may differ from the one specified in the settings
    final_time_sim = my_problem.t.value

    def exact_solution_end(x):
        return exact_solution(x, final_time_sim)

    L2_error = error_L2(computed_solution, exact_solution_end)
    assert L2_error < 1e-7


def test_MES():
    """Method of Exact Solution test for transient heat transfer
    with thermal cond. k = 2, T = 0 on surfaces, and source term q = 8 k

    Analytical solution: T(x) = 4 x (1 - x)
    """
    my_problem = F.HeatTransferProblem()

    my_problem.mesh = F.Mesh1D(vertices=np.linspace(0, 1, 3000))
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)
    mat = F.Material(D_0=None, E_D=None)
    mat.thermal_conductivity = 2

    volume_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=mat)
    my_problem.subdomains = [left, right, volume_subdomain]

    my_problem.boundary_conditions = [
        F.FixedTemperatureBC(subdomain=left, value=0),
        F.FixedTemperatureBC(subdomain=right, value=0),
    ]

    my_problem.sources = [
        F.HeatSource(value=8 * mat.thermal_conductivity, volume=volume_subdomain)
    ]

    my_problem.settings = F.Settings(
        atol=1e-10,
        rtol=1e-10,
        transient=False,
    )

    my_problem.initialise()
    my_problem.run()

    computed_solution = my_problem.u

    def analytical_solution(x):
        return 4 * x[0] * (1 - x[0])

    L2_error = error_L2(computed_solution, analytical_solution)
    assert L2_error < 1e-7


# TODO populate this in other tests
def test_sympify():
    def exact_solution(x, t):
        return 2 * x[0] ** 2 + 20 * t

    def density(T):
        return 0.2 * T + 2

    def heat_capacity(T):
        return 0.2 * T + 3

    def thermal_conductivity(T):
        return 0.1 * T + 4

    mms_source_from_sp = source_from_exact_solution(
        exact_solution,
        density=lambda x, t: density(exact_solution(x, t)),
        heat_capacity=lambda x, t: heat_capacity(exact_solution(x, t)),
        thermal_conductivity=lambda x, t: thermal_conductivity(exact_solution(x, t)),
    )

    def mms_source(x, t):
        return mms_source_from_sp((x[0], None, None), t)

    my_problem = F.HeatTransferProblem()

    my_problem.mesh = F.Mesh1D(vertices=np.linspace(2, 3, 2100))
    left = F.SurfaceSubdomain1D(id=1, x=2)
    right = F.SurfaceSubdomain1D(id=2, x=3)
    mat = F.Material(D_0=None, E_D=None)
    mat.thermal_conductivity = thermal_conductivity
    mat.density = density
    mat.heat_capacity = heat_capacity

    my_problem.subdomains = [
        left,
        right,
        F.VolumeSubdomain1D(id=1, borders=[2, 3], material=mat),
    ]

    # NOTE: it's good to check that without the IC the solution is not the exact one
    my_problem.initial_condition = F.InitialTemperature(lambda x: exact_solution(x, 0))

    my_problem.boundary_conditions = [
        F.FixedTemperatureBC(subdomain=left, value=exact_solution),
        F.FixedTemperatureBC(subdomain=right, value=exact_solution),
    ]

    my_problem.sources = [
        F.HeatSource(value=mms_source, volume=my_problem.volume_subdomains[0])
    ]

    my_problem.settings = F.Settings(
        atol=1e-8,
        rtol=1e-10,
        final_time=1,  # final time shouldn't be too long so that a potential error at the initial timestep is not negligible
    )

    # Forward euler isn't great so dt should be small
    # although it's ok here since the time derivative is constant
    my_problem.settings.stepsize = F.Stepsize(0.05)

    my_problem.exports = [
        F.VTXTemperatureExport(filename="test_transient_heat_transfer.bp")
    ]

    my_problem.initialise()
    my_problem.run()

    computed_solution = my_problem.u
    # we use the exact final time of the simulation which may differ from the one specified in the settings
    final_time_sim = my_problem.t.value

    def exact_solution_end(x):
        return exact_solution(x, final_time_sim)

    L2_error = error_L2(computed_solution, exact_solution_end)
    assert L2_error < 1e-7


def test_sources():
    """Tests the sources setter of the HeatTransferProblem class"""
    htp = F.HeatTransferProblem()
    vol = F.VolumeSubdomain1D(1, borders=[0, 1], material=None)
    # Test that setting valid sources works
    valid_sources = [
        F.HeatSource(value=1, volume=vol),
        F.HeatSource(value=1, volume=vol),
    ]
    htp.sources = valid_sources
    assert htp.sources == valid_sources

    # Test that setting invalid sources raises a TypeError
    with pytest.raises(TypeError, match="festim.HeatSource objects"):
        spe = F.Species("H")
        htp.sources = [
            F.ParticleSource(1, vol, spe),
            F.ParticleSource(1, vol, spe),
        ]


def test_boundary_conditions():
    """Tests the boundary_conditions setter of the HeatTransferProblem class"""
    htp = F.HeatTransferProblem()
    left = F.SurfaceSubdomain(1)

    # Test that setting valid boundary conditions works
    valid_bcs = [F.FixedTemperatureBC(left, 0), F.FixedTemperatureBC(left, 0)]
    htp.boundary_conditions = valid_bcs
    assert htp.boundary_conditions == valid_bcs

    # Test that setting invalid boundary conditions raises a TypeError
    with pytest.raises(TypeError):
        htp.boundary_conditions = [F.FixedConcentrationBC(left, 0, 0)]


mesh_1D = dolfinx.mesh.create_unit_interval(MPI.COMM_WORLD, 10)
mesh_2D = dolfinx.mesh.create_unit_square(MPI.COMM_WORLD, 10, 10)
mesh_3D = dolfinx.mesh.create_unit_cube(MPI.COMM_WORLD, 10, 10, 10)


@pytest.mark.parametrize("mesh", [mesh_1D, mesh_2D, mesh_3D])
def test_meshtags_from_xdmf(tmp_path, mesh):
    """Test that the facet and volume meshtags are read correctly from the mesh XDMF files"""
    # create mesh functions
    fdim = mesh.topology.dim - 1
    vdim = mesh.topology.dim

    # create facet meshtags
    facet_indices = []
    for i in range(vdim):
        # add the boundary entities at 0 and 1 in each dimension
        facets_zero = dolfinx.mesh.locate_entities_boundary(
            mesh, fdim, lambda x: np.isclose(x[i], 0)
        )
        facets_one = dolfinx.mesh.locate_entities_boundary(
            mesh, fdim, lambda x: np.isclose(x[i], 1)
        )

        facet_indices += [facets_zero, facets_one]

    facet_tags = []

    for idx, _ in enumerate(facet_indices):
        # add tags for each boundary
        facet_tag = np.full(len(facet_indices[i]), idx + 1, dtype=np.int32)
        facet_tags.append(facet_tag)

    facet_tags = np.array(facet_tags).flatten()
    facet_indices = np.array(facet_indices).flatten()

    facet_meshtags = dolfinx.mesh.meshtags(mesh, fdim, facet_indices, facet_tags)

    # create volume meshtags
    num_cells = mesh.topology.index_map(vdim).size_local
    mesh_cell_indices = np.arange(num_cells, dtype=np.int32)
    # tag all volumes with 0
    tags_volumes = np.full(num_cells, 0, dtype=np.int32)
    # create 2 volumes for x<0.5 and x>0.5
    volume_indices_left = dolfinx.mesh.locate_entities(
        mesh,
        vdim,
        lambda x: x[0] <= 0.5,
    )

    volume_indices_right = dolfinx.mesh.locate_entities(
        mesh,
        vdim,
        lambda x: x[0] >= 0.5,
    )
    tags_volumes[volume_indices_left] = 2
    tags_volumes[volume_indices_right] = 3

    volume_meshtags = dolfinx.mesh.meshtags(mesh, vdim, mesh_cell_indices, tags_volumes)

    # write files
    surface_file_path = os.path.join(tmp_path, "facets_file.xdmf")
    surface_file = XDMFFile(MPI.COMM_WORLD, surface_file_path, "w")
    surface_file.write_mesh(mesh)
    surface_file.write_meshtags(facet_meshtags, mesh.geometry)

    volume_file_path = os.path.join(tmp_path, "volumes_file.xdmf")
    volume_file = XDMFFile(MPI.COMM_WORLD, volume_file_path, "w")
    volume_file.write_mesh(mesh)
    volume_file.write_meshtags(volume_meshtags, mesh.geometry)

    # read files
    my_model = F.HeatTransferProblem(
        mesh=F.MeshFromXDMF(
            volume_file=volume_file_path,
            facet_file=surface_file_path,
            mesh_name="mesh",
            surface_meshtags_name="mesh_tags",
            volume_meshtags_name="mesh_tags",
        )
    )
    my_model.define_meshtags_and_measures()

    # TEST
    assert volume_meshtags.dim == my_model.volume_meshtags.dim
    assert volume_meshtags.values.all() == my_model.volume_meshtags.values.all()
    assert facet_meshtags.dim == my_model.facet_meshtags.dim
    assert facet_meshtags.values.all() == my_model.facet_meshtags.values.all()


def test_raise_error_non_unique_vol_ids():
    """Test that an error is raised if the volume ids are not unique"""
    my_problem = F.HeatTransferProblem()
    my_problem.mesh = F.Mesh1D(vertices=np.linspace(0, 1, 2000))
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)
    mat = F.Material(D_0=None, E_D=None)
    mat.thermal_conductivity = 1

    my_problem.subdomains = [
        left,
        right,
        F.VolumeSubdomain1D(id=1, borders=[0, 0.5], material=mat),
        F.VolumeSubdomain1D(id=1, borders=[0.5, 1], material=mat),
    ]

    # read files
    with pytest.raises(ValueError):
        my_problem.define_meshtags_and_measures()


def test_initial_condition():
    """Test that an error is raised when initial conditions are defined for a transient simulation"""
    my_problem = F.HeatTransferProblem()

    my_problem.mesh = F.Mesh1D(vertices=np.linspace(0, 1, 3000))
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)
    mat = F.Material(D_0=None, E_D=None)
    mat.thermal_conductivity = 2

    volume_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=mat)
    my_problem.subdomains = [left, right, volume_subdomain]

    my_problem.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

    my_problem.initial_condition = F.InitialTemperature(100)

    with pytest.raises(
        ValueError,
        match="Initial conditions can only be defined for transient simulations",
    ):
        my_problem.initialise()


test_mesh = F.Mesh1D(vertices=np.array([0.0, 1.0, 2.0, 3.0, 4.0]))
dummy_mat = F.Material(D_0=1, E_D=1, name="dummy_mat")
dummy_mat.thermal_conductivity = 1
dummy_mat.density = 1
dummy_mat.heat_capacity = 1


def test_adaptive_timestepping_grows():
    """Tests that the stepsize grows"""
    # BUILD
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    my_model = F.HeatTransferProblem(
        mesh=test_mesh,
        settings=F.Settings(atol=1e-10, rtol=1e-10, transient=True, final_time=10),
        subdomains=[my_vol],
    )

    stepsize = F.Stepsize(initial_value=1)
    stepsize.growth_factor = 1.2
    stepsize.target_nb_iterations = 100  # force it to always grow
    my_model.settings.stepsize = stepsize

    my_model.initialise()

    my_model.progress_bar = tqdm.autonotebook.tqdm(
        desc="Solving H transport problem",
        total=my_model.settings.final_time,
        unit_scale=True,
    )

    # RUN & TEST
    previous_value = stepsize.initial_value
    for i in range(10):
        my_model.iterate()

        # check that the current value is greater than the previous one
        assert my_model.dt.value > previous_value

        previous_value = float(my_model.dt)


def test_adaptive_timestepping_shrinks():
    """Tests that the stepsize shrinks"""
    # BUILD
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    my_model = F.HeatTransferProblem(
        mesh=test_mesh,
        settings=F.Settings(atol=1e-10, rtol=1e-10, transient=True, final_time=10),
        subdomains=[my_vol],
    )

    stepsize = F.Stepsize(initial_value=1)
    stepsize.cutback_factor = 0.8
    stepsize.target_nb_iterations = -1  # force it to always shrink
    my_model.settings.stepsize = stepsize

    my_model.initialise()

    my_model.progress_bar = tqdm.autonotebook.tqdm(
        desc="Solving H transport problem",
        total=my_model.settings.final_time,
        unit_scale=True,
    )

    # RUN & TEST
    previous_value = stepsize.initial_value
    for i in range(10):
        my_model.iterate()

        # check that the current value is smaller than the previous one
        assert my_model.dt.value < previous_value

        previous_value = float(my_model.dt)


@pytest.mark.parametrize(
    "bc_value, expected_values",
    [
        (lambda t: t, [1.0, 2.0, 3.0]),
        (lambda t: 1.0 + t, [2.0, 3.0, 4.0]),
        (lambda x, t: 1.0 + x[0] + t, [6.0, 7.0, 8.0]),
        (lambda T, t: T + 2 * t, [12.0, 14.0, 16.0]),
        (
            lambda x, t: ufl.conditional(ufl.lt(t, 1.5), 100.0 + x[0], 0.0),
            [104.0, 0.0, 0.0],
        ),
    ],
)
def test_update_time_dependent_values_HeatFluxBC(bc_value, expected_values):
    """Test that time dependent fluxes are updated at each time step,
    and match an expected value"""
    # BUILD
    my_vol = F.VolumeSubdomain1D(id=1, borders=[0, 4], material=dummy_mat)
    surface = F.SurfaceSubdomain1D(id=2, x=0)

    my_model = F.HeatTransferProblem(mesh=test_mesh, subdomains=[my_vol, surface])
    my_model.t = fem.Constant(my_model.mesh.mesh, 0.0)
    dt = fem.Constant(test_mesh.mesh, 1.0)

    my_bc = F.HeatFluxBC(subdomain=surface, value=bc_value)
    my_model.boundary_conditions = [my_bc]

    my_model.define_function_space()

    # dummy contant value to test temperature depedence
    my_model.u = F.as_fenics_constant(value=10, mesh=test_mesh.mesh)

    my_model.define_meshtags_and_measures()
    my_model.create_initial_conditions()
    my_model.create_flux_values_fenics()

    for i in range(3):
        # RUN
        my_model.t.value += dt.value
        my_model.update_time_dependent_values()

        # TEST
        if isinstance(my_model.boundary_conditions[0].value_fenics, fem.Constant):
            computed_value = float(my_model.boundary_conditions[0].value_fenics)
            assert np.isclose(computed_value, expected_values[i])



import dolfinx
import numpy as np

from festim.helpers_discontinuity import transfer_meshtags_to_submesh
from festim.material import Material


class VolumeSubdomain:
    """
    Volume subdomain class

    Args:
        id (int): the id of the volume subdomain
        material (festim.Material): the material assigned to the subdomain
    """

    id: int
    submesh: dolfinx.mesh.Mesh
    submesh_to_mesh: np.ndarray
    parent_mesh: dolfinx.mesh.Mesh
    parent_to_submesh: np.ndarray
    v_map: np.ndarray
    facet_to_parent: np.ndarray
    ft: dolfinx.mesh.MeshTags
    padded: bool
    u: dolfinx.fem.Function
    u_n: dolfinx.fem.Function
    material: Material

    def __init__(self, id, material):
        self.id = id
        self.material = material

    def locate_subdomain_entities(self, mesh: dolfinx.mesh.Mesh):
        """Locates all cells in subdomain borders within domain

        Args:
            mesh (dolfinx.mesh.Mesh): the mesh of the model

        Returns:
            entities (np.array): the entities of the subdomain
        """
        cell_map = mesh.topology.index_map(mesh.topology.dim)
        num_cells_local = cell_map.size_local + cell_map.num_ghosts
        return np.arange(num_cells_local, dtype=np.int32)

    def locate_subdomain_entities_correct(self, cell_tags: dolfinx.mesh.MeshTags):
        """Locates all cells in subdomain borders within domain

        Args:
            cell_tags (dolfinx.mesh.MeshTags): the mesh tags of the model

        Returns:
            entities (np.array): the entities of the subdomain
        """
        return cell_tags.find(self.id)

    def create_subdomain(self, mesh: dolfinx.mesh.Mesh, marker: dolfinx.mesh.MeshTags):
        """
        Creates the following attributes: ``.parent_mesh``, ``.submesh``, ``.submesh_to_mesh``,
        ``.v_map``, ``padded``, and the entity map ``parent_to_submesh``.

        Only used in ``festim.HTransportProblemDiscontinuous``

        Args:
            mesh (dolfinx.mesh.Mesh): the parent mesh
            marker (dolfinx.mesh.MeshTags): the parent volume markers
        """
        assert marker.dim == mesh.topology.dim
        self.parent_mesh = (
            mesh  # NOTE: it doesn't seem like we use this attribute anywhere
        )
        self.submesh, self.submesh_to_mesh, self.v_map = dolfinx.mesh.create_submesh(
            mesh, marker.dim, marker.find(self.id)
        )[0:3]
        num_cells_local = (
            mesh.topology.index_map(marker.dim).size_local
            + mesh.topology.index_map(marker.dim).num_ghosts
        )
        self.parent_to_submesh = np.full(num_cells_local, -1, dtype=np.int32)
        self.parent_to_submesh[self.submesh_to_mesh] = np.arange(
            len(self.submesh_to_mesh), dtype=np.int32
        )
        self.padded = False

    def transfer_meshtag(self, mesh: dolfinx.mesh.Mesh, tag: dolfinx.mesh.MeshTags):
        # Transfer meshtags to submesh
        assert self.submesh is not None, "Need to call create_subdomain first"
        self.ft, self.facet_to_parent = transfer_meshtags_to_submesh(
            mesh, tag, self.submesh, self.v_map, self.submesh_to_mesh
        )


def find_volume_from_id(id: int, volumes: list):
    """Returns the correct volume subdomain object from a list of volume ids
    based on an int

    Args:
        id (int): the id of the volume subdomain
        volumes (list): the list of volumes

    Returns:
        festim.VolumeSubdomain: the volume subdomain object with the correct id

    Raises:
        ValueError: if the volume name is not found in the list of volumes

    """
    for vol in volumes:
        if vol.id == id:
            return vol
    raise ValueError(f"id {id} not found in list of volumes")



import csv

import festim as F


class SurfaceQuantity:
    """Export SurfaceQuantity

    Args:
        field (festim.Species): species for which the surface flux is computed
        surface (festim.SurfaceSubdomain1D): surface subdomain
        filename (str, optional): name of the file to which the surface flux is exported

    Attributes:
        field (festim.Species): species for which the surface flux is computed
        surface (festim.SurfaceSubdomain): surface subdomain
        filename (str): name of the file to which the surface flux is exported
        t (list): list of time values
        data (list): list of values of the surface quantity
    """

    def __init__(self, field, surface, filename: str = None) -> None:
        self.field = field
        self.surface = surface
        self.filename = filename

        self.t = []
        self.data = []
        self._first_time_export = True

    @property
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        if value is None:
            self._filename = None
        elif not isinstance(value, str):
            raise TypeError("filename must be of type str")
        elif not value.endswith(".csv") and not value.endswith(".txt"):
            raise ValueError("filename must end with .csv or .txt")
        self._filename = value

    @property
    def surface(self):
        return self._surface

    @surface.setter
    def surface(self, value):
        if not isinstance(value, (int, F.SurfaceSubdomain)) or isinstance(value, bool):
            raise TypeError("surface should be an int or F.SurfaceSubdomain")
        self._surface = value

    @property
    def field(self):
        return self._field

    @field.setter
    def field(self, value):
        # check that field is festim.Species
        if not isinstance(value, (F.Species, str)):
            raise TypeError("field must be of type festim.Species")

        self._field = value

    def write(self, t):
        """If the filename doesnt exist yet, create it and write the header,
        then append the time and value to the file"""

        if self.filename is not None:
            if self._first_time_export:
                header = ["t(s)", f"{self.title}"]
                with open(self.filename, mode="w+", newline="") as file:
                    writer = csv.writer(file)
                    writer.writerow(header)
                self._first_time_export = False
            with open(self.filename, mode="a", newline="") as file:
                writer = csv.writer(file)
                writer.writerow([t, self.value])



import dolfinx
import numpy as np
import ufl
from dolfinx import fem

import festim as F


class SourceBase:
    """
    Source class

    Args:
        volume (festim.VolumeSubdomain1D): the volume subdomains where the source is applied
        value (float, int, fem.Constant or callable): the value of the soure
        species (festim.Species): the species to which the source is applied

    Attributes:
        volume (festim.VolumeSubdomain1D): the volume subdomains where the source is applied
        value (float, int, fem.Constant or callable): the value of the soure
        species (festim.Species): the species to which the source is applied
        value_fenics (fem.Function or fem.Constant): the value of the source in
            fenics format
        source_expr (fem.Expression): the expression of the source term that is
            used to update the value_fenics
        time_dependent (bool): True if the value of the source is time dependent
        temperature_dependent (bool): True if the value of the source is temperature
            dependent

    Usage:
        >>> from festim import Source
        >>> Source(volume=my_vol, value=1, species="H")
        >>> Source(volume=my_vol, value=lambda x: 1 + x[0], species="H")
        >>> Source(volume=my_vol, value=lambda t: 1 + t, species="H")
        >>> Source(volume=my_vol, value=lambda T: 1 + T, species="H")
        >>> Source(volume=my_vol, value=lambda x, t: 1 + x[0] + t, species="H")
    """

    def __init__(self, value, volume):
        self.value = value
        self.volume = volume

        self.value_fenics = None
        self.source_expr = None

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, value):
        # check that volume is festim.VolumeSubdomain
        if not isinstance(value, F.VolumeSubdomain):
            raise TypeError("volume must be of type festim.VolumeSubdomain")
        self._volume = value

    @property
    def value_fenics(self):
        return self._value_fenics

    @value_fenics.setter
    def value_fenics(self, value):
        if value is None:
            self._value_fenics = value
            return
        if not isinstance(
            value, (fem.Function, fem.Constant, np.ndarray, ufl.core.expr.Expr)
        ):
            raise TypeError(
                f"Value must be a dolfinx.fem.Function, dolfinx.fem.Constant, np.ndarray or a ufl.core.expr.Expr, not {type(value)}"
            )
        self._value_fenics = value

    @property
    def time_dependent(self):
        if self.value is None:
            return False
        if isinstance(self.value, fem.Constant):
            return False
        if callable(self.value):
            arguments = self.value.__code__.co_varnames
            return "t" in arguments
        else:
            return False

    def update(self, t):
        """Updates the source value

        Args:
            t (float): the time
        """
        if callable(self.value):
            arguments = self.value.__code__.co_varnames
            if isinstance(self.value_fenics, fem.Constant) and "t" in arguments:
                self.value_fenics.value = self.value(t=t)


class ParticleSource(SourceBase):
    def __init__(self, value, volume, species):
        self.species = species
        super().__init__(value, volume)

    @property
    def species(self):
        return self._species

    @species.setter
    def species(self, value):
        # check that species is festim.Species or list of festim.Species
        if not isinstance(value, F.Species):
            raise TypeError("species must be of type festim.Species")

        self._species = value

    @property
    def temperature_dependent(self):
        if self.value is None:
            return False
        if isinstance(self.value, fem.Constant):
            return False
        if callable(self.value):
            arguments = self.value.__code__.co_varnames
            return "T" in arguments
        else:
            return False

    def create_value_fenics(self, mesh, temperature, t: fem.Constant):
        """Creates the value of the source as a fenics object and sets it to
        self.value_fenics.
        If the value is a constant, it is converted to a fenics.Constant.
        If the value is a function of t, it is converted to a fenics.Constant.
        Otherwise, it is converted to a ufl Expression

        Args:
            mesh (dolfinx.mesh.Mesh) : the mesh
            temperature (float): the temperature
            t (dolfinx.fem.Constant): the time
        """
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.value, (int, float)):
            self.value_fenics = F.as_fenics_constant(mesh=mesh, value=self.value)

        elif isinstance(self.value, (fem.Function, ufl.core.expr.Expr)):
            self.value_fenics = self.value

        elif callable(self.value):
            arguments = self.value.__code__.co_varnames

            if "t" in arguments and "x" not in arguments and "T" not in arguments:
                # only t is an argument
                if not isinstance(self.value(t=float(t)), (float, int)):
                    raise ValueError(
                        f"self.value should return a float or an int, not {type(self.value(t=float(t)))} "
                    )
                self.value_fenics = F.as_fenics_constant(
                    mesh=mesh, value=self.value(t=float(t))
                )
            else:
                kwargs = {}
                if "t" in arguments:
                    kwargs["t"] = t
                if "x" in arguments:
                    kwargs["x"] = x
                if "T" in arguments:
                    kwargs["T"] = temperature

                self.value_fenics = self.value(**kwargs)


class HeatSource(SourceBase):
    def __init__(self, value, volume):
        super().__init__(value, volume)

    def create_value_fenics(
        self,
        mesh: dolfinx.mesh.Mesh,
        t: fem.Constant,
    ):
        """Creates the value of the source as a fenics object and sets it to
        self.value_fenics.
        If the value is a constant, it is converted to a fenics.Constant.
        If the value is a function of t, it is converted to a fenics.Constant.
        Otherwise, it is converted to a ufl.Expression

        Args:
            mesh (dolfinx.mesh.Mesh) : the mesh
            t (dolfinx.fem.Constant): the time
        """
        x = ufl.SpatialCoordinate(mesh)

        if isinstance(self.value, (int, float)):
            self.value_fenics = F.as_fenics_constant(mesh=mesh, value=self.value)

        elif isinstance(self.value, (fem.Function, ufl.core.expr.Expr)):
            self.value_fenics = self.value

        elif callable(self.value):
            arguments = self.value.__code__.co_varnames

            if "t" in arguments and "x" not in arguments:
                # only t is an argument
                if not isinstance(self.value(t=float(t)), (float, int)):
                    raise ValueError(
                        f"self.value should return a float or an int, not {type(self.value(t=float(t)))} "
                    )
                self.value_fenics = F.as_fenics_constant(
                    mesh=mesh, value=self.value(t=float(t))
                )
            else:
                kwargs = {}
                if "t" in arguments:
                    kwargs["t"] = t
                if "x" in arguments:
                    kwargs["x"] = x
                # TODO could the source be dependend on T? why not?

                self.value_fenics = self.value(**kwargs)



import numpy as np
import numpy.typing as npt
from dolfinx.mesh import Mesh, locate_entities

from festim.subdomain import VolumeSubdomain


class VolumeSubdomain1D(VolumeSubdomain):
    """
    Volume subdomain class for 1D cases

    Args:
        id (int): the id of the volume subdomain
        borders (list of float): the borders of the volume subdomain
        material (festim.Material): the material of the volume subdomain

    Attributes:
        id (int): the id of the volume subdomain
        borders (list of float): the borders of the volume subdomain
        material (festim.Material): the material of the volume subdomain

    Usage:
        >>> vol_subdomain = F.VolumeSubdomain1D(id=1, borders=[0, 1],
        ...     material=F.Material(...))
    """

    def __init__(self, id, borders, material) -> None:
        super().__init__(id, material)
        self.borders = borders

    def locate_subdomain_entities(self, mesh: Mesh) -> npt.NDArray[np.int32]:
        """Locates all cells in subdomain borders within domain

        Args:
            mesh (dolfinx.mesh.Mesh): the mesh of the model

        Returns:
            entities (np.array): the entities of the subdomain
        """
        entities = locate_entities(
            mesh,
            mesh.topology.dim,
            lambda x: np.logical_and(x[0] >= self.borders[0], x[0] <= self.borders[1]),
        )
        return entities



__all__ = [
    "VolumeQuantity",
    "SurfaceQuantity",
    "VTXSpeciesExport",
    "VTXTemperatureExport",
    "XDMFExport",
    "SurfaceFlux",
    "TotalSurface",
    "AverageSurface",
    "AverageVolume",
    "TotalVolume",
]

from .average_surface import AverageSurface
from .average_volume import AverageVolume
from .surface_flux import SurfaceFlux
from .surface_quantity import SurfaceQuantity
from .total_surface import TotalSurface
from .total_volume import TotalVolume
from .volume_quantity import VolumeQuantity
from .vtx import VTXSpeciesExport, VTXTemperatureExport
from .xdmf import XDMFExport



import numpy as np

import festim as F

my_model = F.HTransportProblemDiscontinuous()

interface_1 = 0.5
interface_2 = 0.7

# for some reason if the mesh isn't fine enough then I have a random SEGV error
N = 1500
vertices = np.concatenate(
    [
        np.linspace(0, interface_1, num=N),
        np.linspace(interface_1, interface_2, num=N),
        np.linspace(interface_2, 1, num=N),
    ]
)

my_model.mesh = F.Mesh1D(vertices)

material_left = F.Material(D_0=2.0, E_D=0, K_S_0=2.0, E_K_S=0)
material_mid = F.Material(D_0=2.0, E_D=0, K_S_0=4.0, E_K_S=0)
material_right = F.Material(D_0=2.0, E_D=0, K_S_0=6.0, E_K_S=0)

left_domain = F.VolumeSubdomain1D(
    3, borders=[vertices[0], interface_1], material=material_left
)
middle_domain = F.VolumeSubdomain1D(
    4, borders=[interface_1, interface_2], material=material_mid
)
right_domain = F.VolumeSubdomain1D(
    5, borders=[interface_2, vertices[-1]], material=material_right
)

left_surface = F.SurfaceSubdomain1D(id=1, x=vertices[0])
right_surface = F.SurfaceSubdomain1D(id=2, x=vertices[-1])

# the ids here are arbitrary in 1D, you can put anything as long as it's not the same as the surfaces
# TODO remove mesh and meshtags from these arguments
my_model.interfaces = [
    F.Interface(6, (left_domain, middle_domain)),
    F.Interface(7, (middle_domain, right_domain)),
]

my_model.subdomains = [
    left_domain,
    middle_domain,
    right_domain,
    left_surface,
    right_surface,
]
my_model.surface_to_volume = {
    right_surface: right_domain,
    left_surface: left_domain,
}

H = F.Species("H", mobile=True)
trapped_H = F.Species("H_trapped", mobile=False)
empty_trap = F.ImplicitSpecies(n=0.5, others=[trapped_H])

my_model.species = [H, trapped_H]

for species in my_model.species:
    species.subdomains = my_model.volume_subdomains


my_model.reactions = [
    F.Reaction(
        reactant=[H, empty_trap],
        product=[trapped_H],
        k_0=2,
        E_k=0,
        p_0=0.1,
        E_p=0,
        volume=domain,
    )
    for domain in [left_domain, middle_domain, right_domain]
]

my_model.boundary_conditions = [
    F.DirichletBC(left_surface, value=0.05, species=H),
    F.DirichletBC(right_surface, value=0.2, species=H),
]


my_model.temperature = lambda x: 300 + 100 * x[0]

my_model.settings = F.Settings(atol=None, rtol=1e-5, transient=True, final_time=100)
my_model.settings.stepsize = 1

my_model.exports = [
    F.VTXSpeciesExport(filename=f"u_{subdomain.id}.bp", field=H, subdomain=subdomain)
    for subdomain in my_model.volume_subdomains
] + [
    F.VTXSpeciesExport(
        filename=f"u_t_{subdomain.id}.bp", field=trapped_H, subdomain=subdomain
    )
    for subdomain in my_model.volume_subdomains
]

my_model.initialise()
my_model.run()



import numpy as np
import ufl
from dolfinx import fem

import festim as F

from .tools import error_L2

test_mesh_1d = F.Mesh1D(np.linspace(0, 1, 10000))
x_1d = ufl.SpatialCoordinate(test_mesh_1d.mesh)


def test_heat_transfer_steady_state():
    """
    MMS test with heat transfer simulation
    """

    def u_exact(mod):
        return lambda x: 1 + mod.sin(2 * mod.pi * x[0])

    T_analytical_ufl = u_exact(ufl)
    T_analytical_np = u_exact(np)

    V = fem.functionspace(test_mesh_1d.mesh, ("Lagrange", 1))
    T_solution = fem.Function(V)
    T_solution.interpolate(lambda x: 1 + np.sin(2 * np.pi * x[0]))

    Thermal_conductivity = 1.5

    my_model = F.HeatTransferProblem()
    my_model.mesh = test_mesh_1d
    my_mat = F.Material(
        name="mat", D_0=1, E_D=1, thermal_conductivity=Thermal_conductivity
    )
    vol = F.VolumeSubdomain1D(id=1, borders=[0, 1], material=my_mat)
    left = F.SurfaceSubdomain1D(id=1, x=0)
    right = F.SurfaceSubdomain1D(id=2, x=1)

    my_model.subdomains = [vol, left, right]

    my_model.boundary_conditions = [
        F.FixedTemperatureBC(subdomain=left, value=T_analytical_ufl),
        F.FixedTemperatureBC(subdomain=right, value=T_analytical_ufl),
    ]

    f = -ufl.div(Thermal_conductivity * ufl.grad(T_analytical_ufl(x_1d)))
    my_model.sources = [F.HeatSource(value=f, volume=vol)]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)

    my_model.initialise()
    my_model.run()

    T_computed = my_model.u

    L2_error = error_L2(T_computed, T_analytical_np)

    assert L2_error < 1e-7



from festim.reaction import Reaction as _Reaction
from festim.species import ImplicitSpecies as _ImplicitSpecies
from festim.species import Species as _Species


class Trap(_Species):
    """Trap species class for H transport simulation.

    This class only works for 1 mobile species and 1 trapping level and is
    for convenience, for more details see notes.

    Args:
        name (str, optional): a name given to the trap. Defaults to None.
        mobile_species (_Species): the mobile species to be trapped
        k_0 (float): the trapping rate constant pre-exponential factor (m3 s-1)
        E_k (float): the trapping rate constant activation energy (eV)
        p_0 (float): the detrapping rate constant pre-exponential factor (s-1)
        E_p (float): the detrapping rate constant activation energy (eV)
        volume (F.VolumeSubdomain1D): The volume subdomain where the trap is.


    Attributes:
        name (str, optional): a name given to the trap. Defaults to None.
        mobile_species (_Species): the mobile species to be trapped
        k_0 (float): the trapping rate constant pre-exponential factor (m3 s-1)
        E_k (float): the trapping rate constant activation energy (eV)
        p_0 (float): the detrapping rate constant pre-exponential factor (s-1)
        E_p (float): the detrapping rate constant activation energy (eV)
        volume (F.VolumeSubdomain1D): The volume subdomain where the trap is.
        trapped_concentration (_Species): The immobile trapped concentration
        trap_reaction (_Reaction): The reaction for trapping the mobile conc.
        empty_trap_sites (F.ImplicitSpecies): The implicit species for the empty trap sites

    Usage:
        >>> import festim as F
        >>> trap = F.Trap(name="Trap", species=H, k_0=1.0, E_k=0.2, p_0=0.1, E_p=0.3, volume=my_vol)
        >>> trap.name
        'Trap'
        >>> my_model = F.HydorgenTransportProblem()
        >>> my_model.traps = [trap]

    Notes:
        This convenience class replaces the need to specify an implicit species and
        the associated reaction, thus:

        cm = _Species("mobile")
        my_trap = F.Trap(
            name="trapped",
            mobile_species=cm,
            k_0=1,
            E_k=1,
            p_0=1,
            E_p=1,
            n=1,
            volume=my_vol,
        )
        my_model.species = [cm]
        my_model.traps = [my_trap]

        is equivalent to:

        cm = _Species("mobile")
        ct = _Species("trapped")
        trap_sites = F.ImplicitSpecies(n=1, others=[ct])
        trap_reaction = _Reaction(
            reactant=[cm, trap_sites],
            product=ct,
            k_0=1,
            E_k=1,
            p_0=1,
            E_p=1,
            volume=my_vol,
        )
        my_model.species = [cm, ct]
        my_model.reactions = [trap_reaction]


    """

    def __init__(
        self, name: str, mobile_species, k_0, E_k, p_0, E_p, n, volume
    ) -> None:
        super().__init__(name)
        self.mobile_species = mobile_species
        self.k_0 = k_0
        self.E_k = E_k
        self.p_0 = p_0
        self.E_p = E_p
        self.n = n
        self.volume = volume

        self.trapped_concentration = None
        self.reaction = None

    def create_species_and_reaction(self):
        """create the immobile trapped species object and the reaction for trapping"""
        self.trapped_concentration = _Species(name=self.name, mobile=False)
        self.empty_trap_sites = _ImplicitSpecies(
            n=self.n, others=[self.trapped_concentration]
        )
        self.reaction = _Reaction(
            reactant=[self.mobile_species, self.empty_trap_sites],
            product=self.trapped_concentration,
            k_0=self.k_0,
            E_k=self.E_k,
            p_0=self.p_0,
            E_p=self.E_p,
            volume=self.volume,
        )



import festim.boundary_conditions
from festim.hydrogen_transport_problem import HydrogenTransportProblem
from festim import boundary_conditions, as_fenics_constant
import festim
import festim.species as _species
import ufl
from dolfinx import fem

from typing import List


class HydrogenTransportProblemDiscontinuousChangeVar(HydrogenTransportProblem):
    species: List[_species.Species]

    def initialise(self):
        self.create_species_from_traps()
        self.define_function_spaces()
        self.define_meshtags_and_measures()
        self.assign_functions_to_species()

        self.t = fem.Constant(self.mesh.mesh, 0.0)
        if self.settings.transient:
            # TODO should raise error if no stepsize is provided
            # TODO Should this be an attribute of festim.Stepsize?
            self.dt = as_fenics_constant(
                self.settings.stepsize.initial_value, self.mesh.mesh
            )

        self.create_implicit_species_value_fenics()

        self.define_temperature()
        self.define_boundary_conditions()
        self.create_source_values_fenics()
        self.create_flux_values_fenics()
        self.create_initial_conditions()
        self.create_formulation()
        self.create_solver()
        self.override_post_processing_solution()  # NOTE this is the only difference with parent class
        self.initialise_exports()

    def create_formulation(self):
        """Creates the formulation of the model"""

        self.formulation = 0

        # add diffusion and time derivative for each species
        for spe in self.species:
            u = spe.solution
            u_n = spe.prev_solution
            v = spe.test_function

            for vol in self.volume_subdomains:
                D = vol.material.get_diffusion_coefficient(
                    self.mesh.mesh, self.temperature_fenics, spe
                )
                if spe.mobile:
                    K_S = vol.material.get_solubility_coefficient(
                        self.mesh.mesh, self.temperature_fenics, spe
                    )
                    c = u * K_S
                    c_n = u_n * K_S
                else:
                    c = u
                    c_n = u_n
                if spe.mobile:
                    self.formulation += ufl.dot(D * ufl.grad(c), ufl.grad(v)) * self.dx(
                        vol.id
                    )

                if self.settings.transient:
                    self.formulation += ((c - c_n) / self.dt) * v * self.dx(vol.id)

        for reaction in self.reactions:
            if isinstance(reaction.product, list):
                products = reaction.product
            else:
                products = [reaction.product]

            # hack enforce the concentration attribute of the species for all species
            # to be used in reaction.reaction_term
            for spe in self.species:
                if spe.mobile:
                    K_S = reaction.volume.material.get_solubility_coefficient(
                        self.mesh.mesh, self.temperature_fenics, spe
                    )
                    spe.concentration = spe.solution * K_S

            # reactant
            for reactant in reaction.reactant:
                if isinstance(reactant, festim.species.Species):
                    self.formulation += (
                        reaction.reaction_term(self.temperature_fenics)
                        * reactant.test_function
                        * self.dx(reaction.volume.id)
                    )

            # product
            for product in products:
                self.formulation += (
                    -reaction.reaction_term(self.temperature_fenics)
                    * product.test_function
                    * self.dx(reaction.volume.id)
                )
        # add sources
        for source in self.sources:
            self.formulation -= (
                source.value_fenics
                * source.species.test_function
                * self.dx(source.volume.id)
            )

        # add fluxes
        for bc in self.boundary_conditions:
            if isinstance(bc, boundary_conditions.ParticleFluxBC):
                self.formulation -= (
                    bc.value_fenics
                    * bc.species.test_function
                    * self.ds(bc.subdomain.id)
                )

        # check if each species is defined in all volumes
        if not self.settings.transient:
            for spe in self.species:
                # if species mobile, already defined in diffusion term
                if not spe.mobile:
                    not_defined_in_volume = self.volume_subdomains.copy()
                    for vol in self.volume_subdomains:
                        # check reactions
                        for reaction in self.reactions:
                            if (
                                spe in reaction.product
                            ):  # TODO we probably need this in HydrogenTransportProblem too no?
                                if vol == reaction.volume:
                                    if vol in not_defined_in_volume:
                                        not_defined_in_volume.remove(vol)

                    # add c = 0 to formulation where needed
                    for vol in not_defined_in_volume:
                        self.formulation += (
                            spe.solution * spe.test_function * self.dx(vol.id)
                        )

    def override_post_processing_solution(self):
        # override the post-processing solution c = theta * K_S
        Q0 = fem.functionspace(self.mesh.mesh, ("DG", 0))
        Q1 = fem.functionspace(self.mesh.mesh, ("DG", 1))

        for spe in self.species:
            if not spe.mobile:
                continue
            K_S0 = fem.Function(Q0)
            E_KS = fem.Function(Q0)
            for subdomain in self.volume_subdomains:
                entities = subdomain.locate_subdomain_entities_correct(
                    self.volume_meshtags
                )
                K_S0.x.array[entities] = subdomain.material.get_K_S_0(spe)
                E_KS.x.array[entities] = subdomain.material.get_E_K_S(spe)

            K_S = K_S0 * ufl.exp(-E_KS / (festim.k_B * self.temperature_fenics))

            theta = spe.solution

            spe.dg_expr = fem.Expression(theta * K_S, Q1.element.interpolation_points())
            spe.post_processing_solution = fem.Function(Q1)
            spe.post_processing_solution.interpolate(
                spe.dg_expr
            )  # NOTE: do we need this line since it's in initialise?

    def post_processing(self):
        # need to compute c = theta * K_S
        # this expression is stored in species.dg_expr
        for spe in self.species:
            if not spe.mobile:
                continue
            spe.post_processing_solution.interpolate(spe.dg_expr)

        super().post_processing()

    def create_dirichletbc_form(self, bc: festim.FixedConcentrationBC):
        """Creates a dirichlet boundary condition form

        Args:
            bc (festim.DirichletBC): the boundary condition

        Returns:
            dolfinx.fem.bcs.DirichletBC: A representation of
                the boundary condition for modifying linear systems.
        """
        # create value_fenics
        if not self.multispecies:
            function_space_value = bc.species.sub_function_space
        else:
            function_space_value = bc.species.collapsed_function_space

        # create K_S function
        Q0 = fem.functionspace(self.mesh.mesh, ("DG", 0))
        K_S0 = fem.Function(Q0)
        E_KS = fem.Function(Q0)
        for subdomain in self.volume_subdomains:
            entities = subdomain.locate_subdomain_entities_correct(self.volume_meshtags)
            K_S0.x.array[entities] = subdomain.material.get_K_S_0(bc.species)
            E_KS.x.array[entities] = subdomain.material.get_E_K_S(bc.species)

        K_S = K_S0 * ufl.exp(-E_KS / (festim.k_B * self.temperature_fenics))

        bc.create_value(
            temperature=self.temperature_fenics,
            function_space=function_space_value,
            t=self.t,
            K_S=K_S,
        )

        # get dofs
        if self.multispecies and isinstance(bc.value_fenics, (fem.Function)):
            function_space_dofs = (
                bc.species.sub_function_space,
                bc.species.collapsed_function_space,
            )
        else:
            function_space_dofs = bc.species.sub_function_space

        bc_dofs = bc.define_surface_subdomain_dofs(
            facet_meshtags=self.facet_meshtags,
            function_space=function_space_dofs,
        )

        # create form
        if not self.multispecies and isinstance(bc.value_fenics, (fem.Function)):
            # no need to pass the functionspace since value_fenics is already a Function
            function_space_form = None
        else:
            function_space_form = bc.species.sub_function_space

        form = fem.dirichletbc(
            value=bc.value_fenics,
            dofs=bc_dofs,
            V=function_space_form,
        )

        return form

    def update_time_dependent_values(self):
        super().update_time_dependent_values()

        if self.temperature_time_dependent:
            for bc in self.boundary_conditions:
                if isinstance(bc, boundary_conditions.FixedConcentrationBC):
                    bc.update(self.t)



import numpy as np
import pytest

import festim as F


@pytest.mark.parametrize("growth_factor, target", [(10, 5), (1.2, 2), (1, 1)])
def test_adaptive_stepsize_grows(growth_factor, target):
    """Checks that the stepsize is increased correctly

    Args:
        growth_factor (float): the growth factor
        target (int): the target number of iterations
    """
    my_stepsize = F.Stepsize(initial_value=2)
    my_stepsize.growth_factor = growth_factor
    my_stepsize.target_nb_iterations = target

    current_value = 2
    new_value = my_stepsize.modify_value(
        value=current_value,
        nb_iterations=my_stepsize.target_nb_iterations - 1,
    )

    expected_value = current_value * my_stepsize.growth_factor
    assert np.isclose(new_value, expected_value)


@pytest.mark.parametrize("cutback_factor, target", [(0.8, 5), (0.5, 2), (1, 1)])
def test_adaptive_stepsize_shrinks(cutback_factor, target):
    """Checks that the stepsize is shrinks correctly

    Args:
        cutback_factor (float): the cutback factor
        target (int): the target number of iterations
    """
    my_stepsize = F.Stepsize(initial_value=2)
    my_stepsize.cutback_factor = cutback_factor
    my_stepsize.target_nb_iterations = target

    current_value = 2
    new_value = my_stepsize.modify_value(
        value=current_value,
        nb_iterations=my_stepsize.target_nb_iterations + 1,
    )

    expected_value = current_value * my_stepsize.cutback_factor
    assert np.isclose(new_value, expected_value)


@pytest.mark.parametrize("nb_its, target", [(1, 4), (5, 4), (4, 4)])
def test_max_stepsize(nb_its, target):
    """Checks that the stepsize is capped at max
    stepsize.

    Args:
        nb_its (int): the current number of iterations
        target (int): the target number of iterations
    """

    my_stepsize = F.Stepsize(initial_value=1)
    my_stepsize.max_stepsize = 4
    my_stepsize.growth_factor = 1.1
    my_stepsize.cutback_factor = 0.9
    my_stepsize.target_nb_iterations = target

    current_value = 10
    new_value = my_stepsize.modify_value(
        value=current_value,
        nb_iterations=nb_its,
    )

    expected_value = my_stepsize.max_stepsize
    assert new_value == expected_value


def test_stepsize_is_unchanged():
    """
    Checks that the stepsize is unchanged when reaches
    the target nb iterations
    """
    my_stepsize = F.Stepsize(initial_value=2)
    my_stepsize.target_nb_iterations = 5

    current_value = 2
    new_value = my_stepsize.modify_value(value=current_value, nb_iterations=5)

    assert np.isclose(new_value, current_value)


def test_custom_stepsize_not_adaptive():
    """Checks that a custom stepsize that isn't adaptive is unchanged"""

    class CustomStepsize(F.Stepsize):
        def is_adapt(self, t):
            return False

    my_stepsize = CustomStepsize(initial_value=2)
    current_value = 2
    new_value = my_stepsize.modify_value(value=current_value, nb_iterations=5)

    assert np.isclose(new_value, current_value)


def test_growth_factor_setter():
    """Checks that the growth factor setter works correctly"""
    stepsize = F.Stepsize(1)

    # Test that setting a growth factor less than 1 raises a ValueError
    with pytest.raises(ValueError, match="growth factor should be greater than one"):
        stepsize.growth_factor = 0.5

    # Test that setting growth factor to None works
    stepsize.growth_factor = None
    assert stepsize.growth_factor is None


def test_cutback_factor_setter():
    """Checks that the cutback factor setter works correctly"""
    stepsize = F.Stepsize(1)

    # Test that setting a cutback factor greater than 1 raises a ValueError
    with pytest.raises(ValueError, match="cutback factor should be smaller than one"):
        stepsize.cutback_factor = 1.5

    # Test that setting cutback factor to None works
    stepsize.cutback_factor = None
    assert stepsize.cutback_factor is None


def test_max_stepsize_setter():
    """Checks that the maximum stepsize setter works correctly"""
    stepsize = F.Stepsize(initial_value=1)

    # Test that setting a maximum stepsize less than initial stepsize raises a ValueError
    with pytest.raises(
        ValueError, match="maximum stepsize cannot be less than initial stepsize"
    ):
        stepsize.max_stepsize = 0.5

    # Test that setting maximum stepsize to None works
    stepsize.max_stepsize = None
    assert stepsize.max_stepsize is None


@pytest.mark.parametrize(
    "milestones, current_time, expected_value",
    [
        ([1.0, 25.4], 20.1, 25.4),
        ([9.8], 10.0, None),
        ([2.0, 0.5, 20.0], 0.0, 0.5),
        ([3.4, 9.5, 4.4], 4.4, 9.5),
        ([15.3, 1.2, 0.7, 1.4], 15.3, None),
    ],
)
def test_next_milestone(milestones, current_time, expected_value):
    """Checks that the next milestone is
     identified and set correctly.

    Args:
        milestone (float): next milestone
        current_time (float): current time in simulation
    """
    stepsize = F.Stepsize(initial_value=0.5)

    stepsize.milestones = milestones

    next_milestone = stepsize.next_milestone(current_time=current_time)
    assert expected_value == next_milestone


def test_overshoot_milestone():
    """Test that stepsize is modified
    when going to overshoot a milestone.
    """

    my_stepsize = F.Stepsize(initial_value=0.1)
    my_stepsize.growth_factor = 1
    my_stepsize.target_nb_iterations = 4

    my_stepsize.milestones = [1.3]

    current_value = 100000
    new_value = my_stepsize.modify_value(value=current_value, nb_iterations=1, t=1)

    expected_value = 1.3 - 1

    assert new_value == expected_value


@pytest.mark.parametrize(
    "t, expected_value",
    [(0, 10), (10, 10), (100, 10), (1000, None), (1001, None)],
)
def test_get_max_stepsize(t, expected_value):
    """Tests get_max_stepsize when
    max_stepsize is a callable.
    """
    my_stepsize = F.Stepsize(initial_value=2)

    my_stepsize.max_stepsize = lambda t: 10 if t < 1000 else None

    assert my_stepsize.get_max_stepsize(t) == expected_value



from mpi4py import MPI

import dolfinx
import dolfinx.fem.petsc
import numpy as np
import ufl

import festim as F

from .tools import error_L2


def generate_mesh(n=20):
    def bottom_boundary(x):
        return np.isclose(x[1], 0.0)

    def top_boundary(x):
        return np.isclose(x[1], 1.0)

    def half(x):
        return x[1] <= 0.5 + 1e-14

    mesh = dolfinx.mesh.create_unit_square(
        MPI.COMM_WORLD, n, n, dolfinx.mesh.CellType.triangle
    )

    # Split domain in half and set an interface tag of 5
    gdim = mesh.geometry.dim
    tdim = mesh.topology.dim
    fdim = tdim - 1
    top_facets = dolfinx.mesh.locate_entities_boundary(mesh, fdim, top_boundary)
    bottom_facets = dolfinx.mesh.locate_entities_boundary(mesh, fdim, bottom_boundary)
    num_facets_local = (
        mesh.topology.index_map(fdim).size_local
        + mesh.topology.index_map(fdim).num_ghosts
    )
    facets = np.arange(num_facets_local, dtype=np.int32)
    values = np.full_like(facets, 0, dtype=np.int32)
    values[top_facets] = 1
    values[bottom_facets] = 2

    bottom_cells = dolfinx.mesh.locate_entities(mesh, tdim, half)
    num_cells_local = (
        mesh.topology.index_map(tdim).size_local
        + mesh.topology.index_map(tdim).num_ghosts
    )
    cells = np.full(num_cells_local, 4, dtype=np.int32)
    cells[bottom_cells] = 3
    ct = dolfinx.mesh.meshtags(
        mesh, tdim, np.arange(num_cells_local, dtype=np.int32), cells
    )
    all_b_facets = dolfinx.mesh.compute_incident_entities(
        mesh.topology, ct.find(3), tdim, fdim
    )
    all_t_facets = dolfinx.mesh.compute_incident_entities(
        mesh.topology, ct.find(4), tdim, fdim
    )
    interface = np.intersect1d(all_b_facets, all_t_facets)
    values[interface] = 5

    mt = dolfinx.mesh.meshtags(mesh, mesh.topology.dim - 1, facets, values)
    return mesh, mt, ct


def test_2_materials_2d_mms():
    """
    MMS case for a 2D problem with 2 materials
    adapted from https://festim-vv-report.readthedocs.io/en/v1.0/verification/mms/discontinuity.html
    """
    K_S_top = 3.0
    K_S_bot = 6.0
    D_top = 2.0
    D_bot = 5.0
    c_exact_top_ufl = (
        lambda x: 3
        + ufl.sin(ufl.pi * (2 * x[1] + 0.5))
        + 0.1 * ufl.cos(2 * ufl.pi * x[0])
    )
    c_exact_top_np = (
        lambda x: 3 + np.sin(np.pi * (2 * x[1] + 0.5)) + 0.1 * np.cos(2 * np.pi * x[0])
    )

    def c_exact_bot_ufl(x):
        return K_S_bot / K_S_top**2 * c_exact_top_ufl(x) ** 2

    def c_exact_bot_np(x):
        return K_S_bot / K_S_top**2 * c_exact_top_np(x) ** 2

    mesh, mt, ct = generate_mesh(100)

    my_model = F.HTransportProblemPenalty()
    my_model.mesh = F.Mesh(mesh)
    my_model.volume_meshtags = ct
    my_model.facet_meshtags = mt

    material_top = F.Material(D_0=D_top, E_D=0, K_S_0=K_S_top, E_K_S=0)
    material_bottom = F.Material(D_0=D_bot, E_D=0, K_S_0=K_S_bot, E_K_S=0)

    material_top.solubility_law = "sievert"
    material_bottom.solubility_law = "henry"

    top_domain = F.VolumeSubdomain(4, material=material_top)
    bottom_domain = F.VolumeSubdomain(3, material=material_bottom)

    top_surface = F.SurfaceSubdomain(id=1)
    bottom_surface = F.SurfaceSubdomain(id=2)
    my_model.subdomains = [
        bottom_domain,
        top_domain,
        top_surface,
        bottom_surface,
    ]

    my_model.interfaces = [
        F.Interface(5, (bottom_domain, top_domain), penalty_term=1),
    ]
    my_model.surface_to_volume = {
        top_surface: top_domain,
        bottom_surface: bottom_domain,
    }

    H = F.Species("H", mobile=True)

    my_model.species = [H]

    for species in my_model.species:
        species.subdomains = [bottom_domain, top_domain]

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(top_surface, value=c_exact_top_ufl, species=H),
        F.FixedConcentrationBC(bottom_surface, value=c_exact_bot_ufl, species=H),
    ]

    x = ufl.SpatialCoordinate(mesh)

    source_top_val = -ufl.div(D_top * ufl.grad(c_exact_top_ufl(x)))
    source_bottom_val = -ufl.div(D_bot * ufl.grad(c_exact_bot_ufl(x)))
    my_model.sources = [
        F.ParticleSource(volume=top_domain, species=H, value=source_top_val),
        F.ParticleSource(volume=bottom_domain, species=H, value=source_bottom_val),
    ]

    my_model.temperature = 500.0  # lambda x: 300 + 10 * x[1] + 100 * x[0]

    my_model.settings = F.Settings(atol=1e-10, rtol=1e-10, transient=False)
    my_model.exports = [
        F.VTXSpeciesExport(f"u_{subdomain.id}.bp", field=H, subdomain=subdomain)
        for subdomain in my_model.volume_subdomains
    ]

    my_model.initialise()
    my_model.run()

    c_top_computed = H.subdomain_to_post_processing_solution[top_domain]
    c_bot_computed = H.subdomain_to_post_processing_solution[bottom_domain]

    L2_error_top = error_L2(c_top_computed, c_exact_top_np)
    L2_error_bot = error_L2(c_bot_computed, c_exact_bot_np)

    assert L2_error_top < 1e-3
    assert L2_error_bot < 1e-3



import dolfinx
import festim as F
import numpy as np


def test_petsc_options():
    my_model = F.HydrogenTransportProblem(
        petsc_options={
            "ksp_type": "preonly",
            "pc_type": "lu",
            "pc_factor_mat_solver_type": "mumps",
            "coucoucou": 3,
        }
    )

    tungsten = F.Material(D_0=1, E_D=0, K_S_0=1, E_K_S=0)

    my_model.mesh = F.Mesh1D(vertices=np.linspace(0, 1, num=100))
    vol1 = F.VolumeSubdomain1D(id=1, material=tungsten, borders=[0, 1])
    surface1 = F.SurfaceSubdomain1D(id=4, x=0)
    surface2 = F.SurfaceSubdomain1D(id=5, x=1)

    my_model.subdomains = [vol1, surface1, surface2]

    mobile = F.Species(name="H", mobile=True)

    my_model.species = [mobile]

    my_model.temperature = 600

    my_model.boundary_conditions = [
        F.FixedConcentrationBC(subdomain=surface1, species=mobile, value=1),
        F.FixedConcentrationBC(subdomain=surface2, species=mobile, value=0),
    ]

    my_model.settings = F.Settings(atol=1e-6, rtol=1e-6, transient=False)

    my_model.initialise()
    my_model.run()



import numpy as np

import festim.exports.surface_quantity as sq


class MinimumSurface(sq.SurfaceQuantity):
    """Computes the minimum value of a field on a given surface

    Args:
        field (festim.Species): species for which the minimum surface is computed
        surface (festim.SurfaceSubdomain): surface subdomain
        filename (str, optional): name of the file to which the minimum surface is exported

    Attributes:
        see `festim.SurfaceQuantity`
    """

    @property
    def title(self):
        return f"Minimum {self.field.name} surface {self.surface.id}"

    def compute(self):
        """
        Computes the minimum value of the field on the defined surface
        subdomain, and appends it to the data list
        """
        solution = self.field.solution
        indices = self.surface.locate_boundary_facet_indices(
            solution.function_space.mesh
        )
        self.value = np.min(self.field.solution.x.array[indices])
        self.data.append(self.value)
